import sys
from datetime import datetime
from typing import List, Optional, TypeVar
from unittest import mock
from uuid import uuid4

import pytest
import pytz
from pydantic import Field, ValidationError, conlist
from pydantic.main import BaseModel
from pydantic.types import ConstrainedList
from sqlalchemy.exc import NoResultFound

from orchestrator.db import (
    ProductTable,
    SubscriptionInstanceRelationTable,
    SubscriptionInstanceTable,
    SubscriptionInstanceValueTable,
    db,
)
from orchestrator.domain import SUBSCRIPTION_MODEL_REGISTRY
from orchestrator.domain.base import (
    DomainModel,
    ProductBlockModel,
    SubscriptionInstanceList,
    SubscriptionModel,
    _is_constrained_list_type,
    serializable_property,
)
from orchestrator.domain.lifecycle import ProductLifecycle
from orchestrator.types import SubscriptionLifecycle
from test.unit_tests.fixtures.products.product_blocks.product_block_list_nested import (
    ProductBlockListNestedForTestInactive,
)
from test.unit_tests.fixtures.products.product_blocks.product_block_one_nested import (
    ProductBlockOneNestedForTestInactive,
)


def test_product_block_metadata(test_product_block_one, test_product_one, test_product_block_one_db):
    ProductBlockOneForTestInactive, _, _ = test_product_block_one

    subscription_id = uuid4()
    ProductBlockOneForTestInactive.new(
        subscription_id=subscription_id
    )  # Need at least one instance since we lazy load this stuff

    product_block, _ = test_product_block_one_db

    assert ProductBlockOneForTestInactive.name == "ProductBlockOneForTest"
    assert ProductBlockOneForTestInactive.description == "Test Block"
    assert ProductBlockOneForTestInactive.product_block_id == product_block.product_block_id
    assert ProductBlockOneForTestInactive.tag == "TEST"


def test_product_block_one_nested(test_product_model_nested, test_product_type_one_nested):
    """Test the behavior of nesting (self-referencing) product blocks.

    Notes:
        - nesting only works when each block is attached to a different subscription

    """
    ProductTypeOneNestedForTestInactive, _, ProductTypeOneNestedForTest = test_product_type_one_nested

    customer_id = str(uuid4())
    # Create productblock 30 that will be nested in block 20
    model30 = ProductTypeOneNestedForTestInactive.from_product_id(
        product_id=test_product_model_nested.product_id,
        customer_id=customer_id,
        insync=True,
        start_date=None,
        status=SubscriptionLifecycle.INITIAL,
    )
    model30.block = ProductBlockOneNestedForTestInactive.new(
        subscription_id=model30.subscription_id,
        int_field=30,
        sub_block=None,
    )
    model30 = SubscriptionModel.from_other_lifecycle(model30, SubscriptionLifecycle.ACTIVE)
    model30.save()
    db.session.commit()

    # Create productblock 20 that refers to block 30, and will be nested in block 10
    model20 = ProductTypeOneNestedForTestInactive.from_product_id(
        product_id=test_product_model_nested.product_id,
        customer_id=customer_id,
        insync=True,
        start_date=None,
        status=SubscriptionLifecycle.INITIAL,
    )
    model20.block = ProductBlockOneNestedForTestInactive.new(
        subscription_id=model20.subscription_id,
        int_field=20,
        sub_block=model30.block,
    )
    model20 = SubscriptionModel.from_other_lifecycle(model20, SubscriptionLifecycle.ACTIVE)
    model20.save()
    db.session.commit()

    # Create productblock 10 that refers to block 20
    model10 = ProductTypeOneNestedForTestInactive.from_product_id(
        product_id=test_product_model_nested.product_id,
        customer_id=customer_id,
        insync=True,
        start_date=None,
        status=SubscriptionLifecycle.INITIAL,
    )
    model10.block = ProductBlockOneNestedForTestInactive.new(
        subscription_id=model10.subscription_id, int_field=10, sub_block=model20.block
    )
    model10 = SubscriptionModel.from_other_lifecycle(model10, SubscriptionLifecycle.ACTIVE)
    model10.save()
    db.session.commit()

    # Load block 10 and verify the nested blocks
    newmodel10 = ProductTypeOneNestedForTest.from_subscription(model10.subscription_id)
    assert newmodel10.block.int_field == 10
    assert newmodel10.block.sub_block.int_field == 20
    assert newmodel10.block.sub_block.sub_block.int_field == 30
    assert newmodel10.block.sub_block.sub_block.sub_block is None

    # Load block 20 and verify the nested block
    newmodel20 = ProductTypeOneNestedForTest.from_subscription(model20.subscription_id)
    assert newmodel20.block.int_field == 20
    assert newmodel20.block.sub_block.int_field == 30
    assert newmodel20.block.sub_block.sub_block is None

    # Create productblock 11 that also refers to block 20
    model11 = ProductTypeOneNestedForTestInactive.from_product_id(
        product_id=test_product_model_nested.product_id,
        customer_id=customer_id,
        insync=True,
        start_date=None,
        status=SubscriptionLifecycle.INITIAL,
    )
    model11.block = ProductBlockOneNestedForTestInactive.new(
        subscription_id=model11.subscription_id, int_field=11, sub_block=model20.block
    )
    model11 = SubscriptionModel.from_other_lifecycle(model11, SubscriptionLifecycle.ACTIVE)
    model11.save()
    db.session.commit()

    # Load block 11 and verify the nested blocks
    newmodel11 = ProductTypeOneNestedForTest.from_subscription(model11.subscription_id)
    assert newmodel11.block.int_field == 11
    assert newmodel11.block.sub_block.int_field == 20
    assert newmodel11.block.sub_block.sub_block.int_field == 30
    assert newmodel11.block.sub_block.sub_block.sub_block is None

    # (again) Load block 10 and verify the nested blocks are same as before
    newmodel10 = ProductTypeOneNestedForTest.from_subscription(model10.subscription_id)
    assert newmodel10.block.int_field == 10
    assert newmodel10.block.sub_block.int_field == 20
    assert newmodel10.block.sub_block.sub_block.int_field == 30
    assert newmodel10.block.sub_block.sub_block.sub_block is None

    # Below part might not be interesting to test, or better off in a separate testcase.
    # I was just curious what happens when we delete things.

    # Remove block 20 from block 10
    model10.block.sub_block = None
    model10.save()
    db.session.commit()

    # Load block 10 and verify the nested block is removed
    newmodel10 = ProductTypeOneNestedForTest.from_subscription(model10.subscription_id)
    assert newmodel10.block.int_field == 10
    assert newmodel10.block.sub_block is None

    # Load block 11 and verify the nested blocks still exist
    newmodel11 = ProductTypeOneNestedForTest.from_subscription(model11.subscription_id)
    assert newmodel11.block.int_field == 11
    assert newmodel11.block.sub_block.int_field == 20
    assert newmodel11.block.sub_block.sub_block.int_field == 30
    assert newmodel11.block.sub_block.sub_block.sub_block is None


def test_product_block_list_nested(test_product_model_list_nested, test_product_type_list_nested):
    """Test the behavior of nesting (self-referencing) a list of product blocks.

    Notes:
        - nesting only works when each block is attached to a different subscription

    """
    ProductTypeListNestedForTestInactive, _, ProductTypeListNestedForTest = test_product_type_list_nested

    customer_id = str(uuid4())
    # Create productblocks 30 and 31 that will both be nested in block 20 and 21
    model30 = ProductTypeListNestedForTestInactive.from_product_id(
        product_id=test_product_model_list_nested.product_id,
        customer_id=customer_id,
        insync=True,
        start_date=None,
        status=SubscriptionLifecycle.INITIAL,
    )
    model30.block = ProductBlockListNestedForTestInactive.new(
        subscription_id=model30.subscription_id,
        int_field=30,
        sub_block_list=[],
    )
    model30 = SubscriptionModel.from_other_lifecycle(model30, SubscriptionLifecycle.ACTIVE)
    model30.save()
    model31 = ProductTypeListNestedForTestInactive.from_product_id(
        product_id=test_product_model_list_nested.product_id,
        customer_id=customer_id,
        insync=True,
        start_date=None,
        status=SubscriptionLifecycle.INITIAL,
    )
    model31.block = ProductBlockListNestedForTestInactive.new(
        subscription_id=model31.subscription_id,
        int_field=31,
        sub_block_list=[],
    )
    model31 = SubscriptionModel.from_other_lifecycle(model31, SubscriptionLifecycle.ACTIVE)
    model31.save()
    db.session.commit()

    # Create productblocks 20 and 21 that both
    # - refer to blocks 30 and 31
    # - will be nested in blocks 10 and 11
    model20 = ProductTypeListNestedForTestInactive.from_product_id(
        product_id=test_product_model_list_nested.product_id,
        customer_id=customer_id,
        insync=True,
        start_date=None,
        status=SubscriptionLifecycle.INITIAL,
    )
    model20.block = ProductBlockListNestedForTestInactive.new(
        subscription_id=model20.subscription_id,
        int_field=20,
        sub_block_list=[model30.block, model31.block],
    )
    model20 = SubscriptionModel.from_other_lifecycle(model20, SubscriptionLifecycle.ACTIVE)
    model20.save()
    model21 = ProductTypeListNestedForTestInactive.from_product_id(
        product_id=test_product_model_list_nested.product_id,
        customer_id=customer_id,
        insync=True,
        start_date=None,
        status=SubscriptionLifecycle.INITIAL,
    )
    model21.block = ProductBlockListNestedForTestInactive.new(
        subscription_id=model21.subscription_id,
        int_field=21,
        sub_block_list=[model30.block, model31.block],
    )
    model21 = SubscriptionModel.from_other_lifecycle(model21, SubscriptionLifecycle.ACTIVE)
    model21.save()
    db.session.commit()

    # Create productblocks 10 and 11 that both refer to blocks 20 and 21
    model10 = ProductTypeListNestedForTestInactive.from_product_id(
        product_id=test_product_model_list_nested.product_id,
        customer_id=customer_id,
        insync=True,
        start_date=None,
        status=SubscriptionLifecycle.INITIAL,
    )
    model10.block = ProductBlockListNestedForTestInactive.new(
        subscription_id=model10.subscription_id,
        int_field=10,
        sub_block_list=[model20.block, model21.block],
    )
    model10 = SubscriptionModel.from_other_lifecycle(model10, SubscriptionLifecycle.ACTIVE)
    model10.save()
    model11 = ProductTypeListNestedForTestInactive.from_product_id(
        product_id=test_product_model_list_nested.product_id,
        customer_id=customer_id,
        insync=True,
        start_date=None,
        status=SubscriptionLifecycle.INITIAL,
    )
    model11.block = ProductBlockListNestedForTestInactive.new(
        subscription_id=model11.subscription_id,
        int_field=11,
        sub_block_list=[model20.block, model21.block],
    )
    model11 = SubscriptionModel.from_other_lifecycle(model11, SubscriptionLifecycle.ACTIVE)
    model11.save()
    db.session.commit()

    # Load blocks 10 and 11 and verify the nested blocks
    newmodel10 = ProductTypeListNestedForTest.from_subscription(model10.subscription_id)
    newmodel11 = ProductTypeListNestedForTest.from_subscription(model11.subscription_id)
    assert newmodel10.block.int_field == 10
    assert newmodel11.block.int_field == 11
    assert newmodel10.block.sub_block_list[0].int_field == 20
    assert newmodel10.block.sub_block_list[1].int_field == 21
    assert newmodel11.block.sub_block_list[0].int_field == 20
    assert newmodel11.block.sub_block_list[1].int_field == 21
    assert sorted(
        level3.int_field for level2 in newmodel10.block.sub_block_list for level3 in level2.sub_block_list
    ) == [30, 30, 31, 31]
    assert sorted(
        level3.int_field for level2 in newmodel11.block.sub_block_list for level3 in level2.sub_block_list
    ) == [30, 30, 31, 31]
    # Assert a few blocks at deepest level to have an empty list
    assert newmodel10.block.sub_block_list[0].sub_block_list[0].sub_block_list == []
    assert newmodel11.block.sub_block_list[1].sub_block_list[1].sub_block_list == []

    # Load block 20 and verify nested blocks
    newmodel20 = ProductTypeListNestedForTest.from_subscription(model20.subscription_id)
    assert newmodel20.block.int_field == 20
    assert newmodel20.block.sub_block_list[0].int_field == 30
    assert newmodel20.block.sub_block_list[1].int_field == 31
    assert newmodel20.block.sub_block_list[0].sub_block_list == []

    # Create productblock 32 and nest it in block 20 (but not 21!)
    model32 = ProductTypeListNestedForTestInactive.from_product_id(
        product_id=test_product_model_list_nested.product_id,
        customer_id=customer_id,
        insync=True,
        start_date=None,
        status=SubscriptionLifecycle.INITIAL,
    )
    model32.block = ProductBlockListNestedForTestInactive.new(
        subscription_id=model30.subscription_id,
        int_field=32,
        sub_block_list=[],
    )
    model32 = SubscriptionModel.from_other_lifecycle(model32, SubscriptionLifecycle.ACTIVE)
    model32.save()
    newmodel20.block.sub_block_list.append(model32.block)
    newmodel20.save()
    db.session.commit()

    # (again) Load blocks 10 and 11 and verify block 32 is present once
    newmodel10 = ProductTypeListNestedForTest.from_subscription(model10.subscription_id)
    newmodel11 = ProductTypeListNestedForTest.from_subscription(model11.subscription_id)
    assert sorted(
        level3.int_field for level2 in newmodel10.block.sub_block_list for level3 in level2.sub_block_list
    ) == [30, 30, 31, 31, 32]
    assert sorted(
        level3.int_field for level2 in newmodel11.block.sub_block_list for level3 in level2.sub_block_list
    ) == [30, 30, 31, 31, 32]


def test_lifecycle(test_product_model, test_product_type_one, test_product_block_one):
    ProductBlockOneForTestInactive, _, _ = test_product_block_one
    ProductTypeOneForTestInactive, _, _ = test_product_type_one
    subscription_id = uuid4()

    # Test create with wrong lifecycle, we can create
    with pytest.raises(ValueError, match=r"is not valid for status active"):
        ProductTypeOneForTestInactive(
            product=test_product_model,
            customer_id=str(uuid4()),
            subscription_id=subscription_id,
            insync=False,
            description="",
            start_date=None,
            end_date=None,
            note=None,
            block=ProductBlockOneForTestInactive.new(subscription_id),
            status=SubscriptionLifecycle.ACTIVE,
            test_fixed_input=False,
        )

    # Works with right lifecycle
    product_type = ProductTypeOneForTestInactive(
        product=test_product_model,
        customer_id=str(uuid4()),
        subscription_id=subscription_id,
        insync=False,
        description="",
        start_date=None,
        end_date=None,
        note=None,
        block=ProductBlockOneForTestInactive.new(subscription_id),
        status=SubscriptionLifecycle.INITIAL,
        test_fixed_input=False,
    )

    assert product_type.status == SubscriptionLifecycle.INITIAL


def test_lifecycle_specific(
    test_product_model, test_product_type_one, test_product_block_one, test_product_sub_block_one
):
    _, SubBlockOneForTestProvisioning, SubBlockOneForTest = test_product_sub_block_one
    _, ProductBlockOneForTestProvisioning, ProductBlockOneForTest = test_product_block_one
    _, ProductTypeOneForTestProvisioning, ProductTypeOneForTest = test_product_type_one
    subscription_id = uuid4()

    # Works with less contrained lifecycle
    product_type = ProductTypeOneForTest(
        product=test_product_model,
        customer_id=str(uuid4()),
        subscription_id=subscription_id,
        insync=False,
        description="",
        start_date=None,
        end_date=None,
        note=None,
        block=ProductBlockOneForTest.new(
            subscription_id=subscription_id,
            int_field=3,
            str_field="",
            list_field=[1],
            sub_block=SubBlockOneForTest.new(subscription_id=subscription_id, int_field=3, str_field=""),
            sub_block_2=SubBlockOneForTest.new(subscription_id=subscription_id, int_field=3, str_field=""),
        ),
        status=SubscriptionLifecycle.INITIAL,
        test_fixed_input=False,
    )

    assert product_type.status == SubscriptionLifecycle.INITIAL

    # Works with right lifecycle
    product_type = ProductTypeOneForTest(
        product=test_product_model,
        customer_id=str(uuid4()),
        subscription_id=subscription_id,
        insync=False,
        description="",
        start_date=None,
        end_date=None,
        note=None,
        block=ProductBlockOneForTest.new(
            subscription_id=subscription_id,
            int_field=3,
            str_field="",
            list_field=[1],
            sub_block=SubBlockOneForTest.new(subscription_id=subscription_id, int_field=3, str_field=""),
            sub_block_2=SubBlockOneForTest.new(subscription_id=subscription_id, int_field=3, str_field=""),
        ),
        status=SubscriptionLifecycle.ACTIVE,
        test_fixed_input=False,
    )

    assert product_type.status == SubscriptionLifecycle.ACTIVE

    # Does not work with more constrained lifecycle
    with pytest.raises(ValueError, match=r"is not valid for status active"):
        ProductTypeOneForTestProvisioning(
            product=test_product_model,
            customer_id=str(uuid4()),
            subscription_id=subscription_id,
            insync=False,
            description="",
            start_date=None,
            end_date=None,
            note=None,
            block=ProductBlockOneForTestProvisioning.new(
                subscription_id=subscription_id,
                int_field=3,
                list_field=[1],
                sub_block=SubBlockOneForTestProvisioning.new(subscription_id=subscription_id, int_field=3),
                sub_block_2=SubBlockOneForTest.new(subscription_id=subscription_id, int_field=3, str_field=""),
            ),
            status=SubscriptionLifecycle.ACTIVE,
            test_fixed_input=False,
        )

    # Works with right lifecycle
    product_type = ProductTypeOneForTestProvisioning(
        product=test_product_model,
        customer_id=str(uuid4()),
        subscription_id=subscription_id,
        insync=False,
        description="",
        start_date=None,
        end_date=None,
        note=None,
        block=ProductBlockOneForTestProvisioning.new(
            subscription_id=subscription_id,
            int_field=3,
            list_field=[1],
            sub_block=SubBlockOneForTestProvisioning.new(subscription_id=subscription_id, int_field=3),
            sub_block_2=SubBlockOneForTest.new(subscription_id=subscription_id, int_field=3, str_field=""),
        ),
        status=SubscriptionLifecycle.PROVISIONING,
        test_fixed_input=False,
    )
    assert product_type.status == SubscriptionLifecycle.PROVISIONING


def test_product_blocks_per_lifecycle(
    test_product_model, test_product_type_one, test_product_block_one, test_product_sub_block_one
):
    _, _, SubBlockOneForTest = test_product_sub_block_one
    ProductBlockOneForTestInactive, _, ProductBlockOneForTest = test_product_block_one
    ProductTypeOneForTestInactive, ProductTypeOneForTestProvisioning, ProductTypeOneForTest = test_product_type_one
    subscription_id = uuid4()

    ProductTypeOneForTestInactive(
        product=test_product_model,
        customer_id=str(uuid4()),
        subscription_id=subscription_id,
        insync=False,
        description="",
        start_date=None,
        end_date=None,
        note=None,
        block=ProductBlockOneForTest.new(
            subscription_id=subscription_id,
            int_field=3,
            str_field="",
            list_field=[1],
            sub_block=SubBlockOneForTest.new(subscription_id=subscription_id, int_field=3, str_field=""),
            sub_block_2=SubBlockOneForTest.new(subscription_id=subscription_id, int_field=3, str_field=""),
        ),
        status=SubscriptionLifecycle.INITIAL,
        test_fixed_input=False,
    )

    ProductTypeOneForTestInactive(
        product=test_product_model,
        customer_id=str(uuid4()),
        subscription_id=subscription_id,
        insync=False,
        description="",
        start_date=None,
        end_date=None,
        note=None,
        block=ProductBlockOneForTestInactive.new(subscription_id=subscription_id, int_field=3),
        status=SubscriptionLifecycle.INITIAL,
        test_fixed_input=False,
    )

    ProductTypeOneForTestProvisioning(
        product=test_product_model,
        customer_id=str(uuid4()),
        subscription_id=subscription_id,
        insync=False,
        description="",
        start_date=None,
        end_date=None,
        note=None,
        block=ProductBlockOneForTest.new(
            subscription_id=subscription_id,
            int_field=3,
            str_field="",
            list_field=[1],
            sub_block=SubBlockOneForTest.new(subscription_id=subscription_id, int_field=3, str_field=""),
            sub_block_2=SubBlockOneForTest.new(subscription_id=subscription_id, int_field=3, str_field=""),
        ),
        status=SubscriptionLifecycle.PROVISIONING,
        test_fixed_input=False,
    )

    ProductTypeOneForTest(
        product=test_product_model,
        customer_id=str(uuid4()),
        subscription_id=subscription_id,
        insync=False,
        description="",
        start_date=None,
        end_date=None,
        note=None,
        block=ProductBlockOneForTest.new(
            subscription_id=subscription_id,
            int_field=3,
            str_field="",
            list_field=[1],
            sub_block=SubBlockOneForTest.new(subscription_id=subscription_id, int_field=3, str_field=""),
            sub_block_2=SubBlockOneForTest.new(subscription_id=subscription_id, int_field=3, str_field=""),
        ),
        status=SubscriptionLifecycle.ACTIVE,
        test_fixed_input=False,
    )

    ProductTypeOneForTest(
        product=test_product_model,
        customer_id=str(uuid4()),
        subscription_id=subscription_id,
        insync=False,
        description="",
        start_date=None,
        end_date=None,
        note=None,
        block=ProductBlockOneForTest.new(
            subscription_id=subscription_id,
            int_field=3,
            str_field="",
            list_field=[1],
            sub_block=SubBlockOneForTest.new(subscription_id=subscription_id, int_field=3, str_field=""),
            sub_block_2=SubBlockOneForTest.new(subscription_id=subscription_id, int_field=3, str_field=""),
        ),
        status=SubscriptionLifecycle.INITIAL,
        test_fixed_input=False,
    )

    with pytest.raises(
        ValidationError,
        match=r"2 validation errors for SubBlockOneForTest\nint_field\n  field required .+\nstr_field\n  field required .+",
    ):
        ProductTypeOneForTest(
            product=test_product_model,
            customer_id=str(uuid4()),
            subscription_id=subscription_id,
            insync=False,
            description="",
            start_date=None,
            end_date=None,
            note=None,
            block=ProductBlockOneForTest.new(subscription_id=subscription_id),
            status=SubscriptionLifecycle.ACTIVE,
            test_fixed_input=False,
        )

    with pytest.raises(ValidationError, match=r"5 validation errors for ProductTypeOneForTest"):
        ProductTypeOneForTest(
            product=test_product_model,
            customer_id=str(uuid4()),
            subscription_id=subscription_id,
            insync=False,
            description="",
            start_date=None,
            end_date=None,
            note=None,
            block=ProductBlockOneForTestInactive.new(subscription_id=subscription_id),
            status=SubscriptionLifecycle.ACTIVE,
            test_fixed_input=False,
        )

    with pytest.raises(
        ValidationError, match=r"1 validation error for ProductTypeOneForTest\nblock\n  field required .+"
    ):
        ProductTypeOneForTest(
            product=test_product_model,
            customer_id=str(uuid4()),
            subscription_id=subscription_id,
            insync=False,
            description="",
            start_date=None,
            end_date=None,
            note=None,
            status=SubscriptionLifecycle.INITIAL,
            test_fixed_input=False,
        )

    with pytest.raises(
        ValidationError, match=r"1 validation error for ProductTypeOneForTest\nblock\n  field required .+"
    ):
        ProductTypeOneForTest(
            product=test_product_model,
            customer_id=str(uuid4()),
            subscription_id=subscription_id,
            insync=False,
            description="",
            start_date=None,
            end_date=None,
            note=None,
            status=SubscriptionLifecycle.ACTIVE,
            test_fixed_input=False,
        )


def test_change_lifecycle(test_product_one, test_product_type_one, test_product_block_one, test_product_sub_block_one):
    SubBlockOneForTestInactive, _, _ = test_product_sub_block_one
    ProductBlockOneForTestInactive, _, _ = test_product_block_one
    ProductTypeOneForTestInactive, _, _ = test_product_type_one

    product_type = ProductTypeOneForTestInactive.from_product_id(
        test_product_one,
        str(uuid4()),
    )
    product_type.block = ProductBlockOneForTestInactive.new(
        subscription_id=product_type.subscription_id,
        sub_block_2=SubBlockOneForTestInactive.new(subscription_id=product_type.subscription_id),
    )

    # Does not work if constraints are not met
    with pytest.raises(ValidationError, match=r"int_field\n  none is not an allowed value"):
        product_type = SubscriptionModel.from_other_lifecycle(product_type, SubscriptionLifecycle.ACTIVE)
    with pytest.raises(ValidationError, match=r"int_field\n  none is not an allowed value"):
        product_type = SubscriptionModel.from_other_lifecycle(product_type, SubscriptionLifecycle.PROVISIONING)

    # Set first value
    product_type.block.int_field = 3
    product_type.block.sub_block.int_field = 3
    product_type.block.sub_block_2.int_field = 3
    product_type.block.sub_block_list = [SubBlockOneForTestInactive.new(subscription_id=product_type.subscription_id)]
    product_type.block.sub_block_list[0].int_field = 4
    product_type.block.list_field = [1]

    # Does not work if constraints are not met
    with pytest.raises(ValidationError, match=r"str_field\n  none is not an allowed value"):
        product_type = SubscriptionModel.from_other_lifecycle(product_type, SubscriptionLifecycle.ACTIVE)

    # works with correct data
    product_type = SubscriptionModel.from_other_lifecycle(product_type, SubscriptionLifecycle.PROVISIONING)
    assert product_type.status == SubscriptionLifecycle.PROVISIONING
    assert product_type.block.str_field is None

    product_type.block.str_field = "A"
    product_type.block.sub_block.str_field = "B"
    product_type.block.sub_block_2.str_field = "C"
    product_type.block.sub_block_list[0].str_field = "D"

    # works with correct data
    product_type_new = SubscriptionModel.from_other_lifecycle(product_type, SubscriptionLifecycle.ACTIVE)
    assert product_type_new.status == SubscriptionLifecycle.ACTIVE
    expected_dict = product_type.dict()
    expected_dict["status"] = SubscriptionLifecycle.ACTIVE
    expected_dict["start_date"] = mock.ANY
    assert product_type_new.dict() == expected_dict
    assert isinstance(product_type_new.start_date, datetime)


def test_save_load(test_product_model, test_product_type_one, test_product_block_one, test_product_sub_block_one):
    SubBlockOneForTestInactive, SubBlockOneForTestProvisioning, SubBlockOneForTest = test_product_sub_block_one
    ProductBlockOneForTestInactive, ProductBlockOneForTestProvisioning, ProductBlockOneForTest = test_product_block_one
    ProductTypeOneForTestInactive, ProductTypeOneForTestProvisioning, ProductTypeOneForTest = test_product_type_one

    customer_id = str(uuid4())

    model = ProductTypeOneForTestInactive.from_product_id(
        product_id=test_product_model.product_id,
        customer_id=customer_id,
        insync=True,
        description="Desc",
        start_date=datetime(2021, 1, 1, 1, 1, 1, tzinfo=pytz.utc),
        end_date=datetime(2021, 1, 1, 1, 1, 2, tzinfo=pytz.utc),
        note="Note",
        status=SubscriptionLifecycle.INITIAL,
    )

    # Set one value and make sure its saved
    model.block.str_field = "A"

    model.save()
    db.session.commit()

    assert (
        SubscriptionInstanceValueTable.query.join(SubscriptionInstanceValueTable.subscription_instance)
        .filter(SubscriptionInstanceTable.subscription_id == model.subscription_id)
        .count()
        == 1
    )

    assert model.dict() == {
        "block": {
            "int_field": None,
            "label": None,
            "list_field": [],
            "name": "ProductBlockOneForTest",
            "str_field": "A",
            "sub_block": {
                "int_field": None,
                "label": None,
                "name": "SubBlockOneForTest",
                "str_field": None,
                "subscription_instance_id": mock.ANY,
                "owner_subscription_id": mock.ANY,
            },
            "sub_block_2": None,
            "sub_block_list": [],
            "owner_subscription_id": mock.ANY,
            "subscription_instance_id": mock.ANY,
        },
        "customer_id": customer_id,
        "description": "Desc",
        "end_date": datetime(2021, 1, 1, 1, 1, 2, tzinfo=pytz.utc),
        "insync": True,
        "note": "Note",
        "product": {
            "description": "Test ProductTable",
            "name": "TestProductOne",
            "product_id": test_product_model.product_id,
            "product_type": "Test",
            "status": ProductLifecycle.ACTIVE,
            "tag": "TEST",
            "created_at": test_product_model.created_at,
            "end_date": None,
        },
        "start_date": datetime(2021, 1, 1, 1, 1, 1, tzinfo=pytz.utc),
        "status": SubscriptionLifecycle.INITIAL,
        "subscription_id": mock.ANY,
        "test_fixed_input": False,
    }

    # Set first value
    model.block.int_field = 3
    model.block.sub_block.int_field = 3
    model.block.sub_block_2 = SubBlockOneForTestInactive.new(subscription_id=model.subscription_id)
    model.block.sub_block_2.int_field = 3
    model.block.list_field = [1]
    model.block.sub_block.str_field = "B"
    model.block.sub_block_2.str_field = "C"

    # works with correct data
    model = SubscriptionModel.from_other_lifecycle(model, SubscriptionLifecycle.ACTIVE)

    model.save()
    db.session.commit()

    new_model = ProductTypeOneForTest.from_subscription(model.subscription_id)
    assert model.dict() == new_model.dict()

    # Second save also works as expected
    new_model.save()
    db.session.commit()

    latest_model = ProductTypeOneForTest.from_subscription(model.subscription_id)
    assert new_model.dict() == latest_model.dict()

    # Loading blocks also works
    block = ProductBlockOneForTest.from_db(model.block.subscription_instance_id)
    assert block.dict() == model.block.dict()


def test_update_constrained_lists(test_product_one, test_product_block_one):
    ProductBlockOneForTestInactive, _, _ = test_product_block_one

    class TestConListProductType(SubscriptionModel, is_base=True):
        saps: conlist(ProductBlockOneForTestInactive, min_items=1, max_items=4)

    # Creates
    ip = TestConListProductType.from_product_id(product_id=test_product_one, customer_id=str(uuid4()))
    ip.save()

    sap = ProductBlockOneForTestInactive.new(subscription_id=ip.subscription_id, int_field=3, str_field="")

    # Set new saps, removes old one
    ip.saps = [sap]

    ip.save()

    ip2 = TestConListProductType.from_subscription(ip.subscription_id)

    # Old default saps should not be saved
    assert ip.dict() == ip2.dict()

    # Test constraint
    with pytest.raises(ValidationError):
        ip.saps = []


def test_update_lists(test_product_one, test_product_block_one):
    ProductBlockOneForTestInactive, _, _ = test_product_block_one

    class TestListProductType(SubscriptionModel, is_base=True):
        saps: List[ProductBlockOneForTestInactive]

    # Creates
    ip = TestListProductType.from_product_id(product_id=test_product_one, customer_id=str(uuid4()))
    ip.save()

    sap = ProductBlockOneForTestInactive.new(subscription_id=ip.subscription_id, int_field=3, str_field="")

    # Set new saps
    ip.saps = [sap]

    ip.save()

    ip2 = TestListProductType.from_subscription(ip.subscription_id)

    # Old default saps should not be saved
    assert ip.dict() == ip2.dict()


def test_update_optional(test_product_one, test_product_block_one):
    ProductBlockOneForTestInactive, _, _ = test_product_block_one

    class TestListProductType(SubscriptionModel, is_base=True):
        sap: Optional[ProductBlockOneForTestInactive] = None

    # Creates
    ip = TestListProductType.from_product_id(product_id=test_product_one, customer_id=str(uuid4()))
    ip.save()

    sap = ProductBlockOneForTestInactive.new(subscription_id=ip.subscription_id, int_field=3, str_field="")

    # Set new sap
    ip.sap = sap

    ip.save()

    ip2 = TestListProductType.from_subscription(ip.subscription_id)

    # Old default saps should not be saved
    assert ip.dict() == ip2.dict()


def test_generic_from_subscription(test_product_one, test_product_type_one):
    ProductTypeOneForTestInactive, _, _ = test_product_type_one

    ip = ProductTypeOneForTestInactive.from_product_id(product_id=test_product_one, customer_id=str(uuid4()))
    ip.save()

    model = SubscriptionModel.from_subscription(ip.subscription_id)

    assert isinstance(model, ProductTypeOneForTestInactive)


def test_label_is_saved(test_product_one, test_product_type_one):
    ProductTypeOneForTestInactive, _, _ = test_product_type_one

    test_model = ProductTypeOneForTestInactive.from_product_id(test_product_one, str(uuid4()))
    test_model.block.label = "My label"
    test_model.save()
    db.session.commit()
    instance_in_db = SubscriptionInstanceTable.query.get(test_model.block.subscription_instance_id)
    assert instance_in_db.label == "My label"


def test_domain_model_attrs_saving_loading(test_product_one, test_product_type_one, test_product_sub_block_one):
    SubBlockOneForTestInactive, _, _ = test_product_sub_block_one
    ProductTypeOneForTestInactive, _, _ = test_product_type_one

    test_model = ProductTypeOneForTestInactive.from_product_id(product_id=test_product_one, customer_id=str(uuid4()))
    test_model.block.sub_block_2 = SubBlockOneForTestInactive.new(subscription_id=test_model.subscription_id)
    test_model.block.sub_block_list = [SubBlockOneForTestInactive.new(subscription_id=test_model.subscription_id)]
    test_model.save()
    db.session.commit()

    relation = SubscriptionInstanceRelationTable.query.filter(
        SubscriptionInstanceRelationTable.depends_on_id == test_model.block.sub_block.subscription_instance_id
    ).one()
    assert relation.domain_model_attr == "sub_block"
    relation = SubscriptionInstanceRelationTable.query.filter(
        SubscriptionInstanceRelationTable.depends_on_id == test_model.block.sub_block_2.subscription_instance_id
    ).one()
    assert relation.domain_model_attr == "sub_block_2"
    relation = SubscriptionInstanceRelationTable.query.filter(
        SubscriptionInstanceRelationTable.depends_on_id == test_model.block.sub_block_list[0].subscription_instance_id
    ).one()
    assert relation.domain_model_attr == "sub_block_list"
    test_model_2 = ProductTypeOneForTestInactive.from_subscription(test_model.subscription_id)
    assert test_model == test_model_2


def test_removal_of_domain_attrs(test_product_one, test_product_type_one, test_product_sub_block_one):
    SubBlockOneForTestInactive, _, _ = test_product_sub_block_one
    ProductTypeOneForTestInactive, _, _ = test_product_type_one

    test_model = ProductTypeOneForTestInactive.from_product_id(product_id=test_product_one, customer_id=str(uuid4()))
    test_model.block.sub_block_2 = SubBlockOneForTestInactive.new(subscription_id=test_model.subscription_id)

    test_model.save()
    db.session.commit()
    relation = SubscriptionInstanceRelationTable.query.filter(
        SubscriptionInstanceRelationTable.depends_on_id == test_model.block.sub_block.subscription_instance_id
    ).one()
    relation.domain_model_attr = None
    db.session.commit()
    with pytest.raises(ValueError, match=r"Expected exactly one item in iterable, but got"):
        ProductTypeOneForTestInactive.from_subscription(test_model.subscription_id)


def test_simple_model_with_no_attrs(generic_subscription_1, generic_product_type_1):
    GenericProductOneInactive, GenericProductOne = generic_product_type_1
    model = GenericProductOne.from_subscription(subscription_id=generic_subscription_1)
    with pytest.raises(NoResultFound):
        SubscriptionInstanceRelationTable.query.filter(
            SubscriptionInstanceRelationTable.depends_on_id == model.pb_1.subscription_instance_id
        ).one()


def test_abstract_super_block(test_product_one, test_product_type_one, test_product_block_one_db):
    ProductTypeOneForTestInactive, ProductTypeOneForTestProvisioning, ProductTypeOneForTest = test_product_type_one

    class AbstractProductBlockOneForTestInactive(ProductBlockModel):
        str_field: Optional[str] = None
        list_field: List[int] = Field(default_factory=list)

    class AbstractProductBlockOneForTestProvisioning(
        AbstractProductBlockOneForTestInactive, lifecycle=[SubscriptionLifecycle.PROVISIONING]
    ):
        str_field: Optional[str] = None
        list_field: List[int]

    class AbstractProductBlockOneForTest(
        AbstractProductBlockOneForTestProvisioning, lifecycle=[SubscriptionLifecycle.ACTIVE]
    ):
        str_field: str
        list_field: List[int]

    class ProductBlockOneForTestInactive(
        AbstractProductBlockOneForTestInactive, product_block_name="ProductBlockOneForTest"
    ):
        str_field: Optional[str] = None
        list_field: List[int] = Field(default_factory=list)
        int_field: Optional[int] = None

    class ProductBlockOneForTestProvisioning(
        ProductBlockOneForTestInactive,
        AbstractProductBlockOneForTestProvisioning,
        lifecycle=[SubscriptionLifecycle.PROVISIONING],
    ):
        str_field: Optional[str] = None
        list_field: List[int]
        int_field: int

    class ProductBlockOneForTest(
        ProductBlockOneForTestProvisioning, AbstractProductBlockOneForTest, lifecycle=[SubscriptionLifecycle.ACTIVE]
    ):
        str_field: str
        list_field: List[int]
        int_field: int

    class AbstractProductTypeOneForTestInactive(SubscriptionModel):
        block: AbstractProductBlockOneForTestInactive

    class AbstractProductTypeOneForTestProvisioning(AbstractProductTypeOneForTestInactive):
        block: AbstractProductBlockOneForTestProvisioning

    class AbstractProductTypeOneForTest(AbstractProductTypeOneForTestProvisioning):
        block: AbstractProductBlockOneForTest

    class ProductTypeOneForTestInactive(AbstractProductTypeOneForTestInactive, is_base=True):
        block: ProductBlockOneForTestInactive

    class ProductTypeOneForTestProvisioning(
        ProductTypeOneForTestInactive,
        AbstractProductTypeOneForTestProvisioning,
        lifecycle=[SubscriptionLifecycle.PROVISIONING],
    ):
        block: ProductBlockOneForTestProvisioning

    class ProductTypeOneForTest(
        ProductTypeOneForTestProvisioning, AbstractProductTypeOneForTest, lifecycle=[SubscriptionLifecycle.ACTIVE]
    ):
        block: ProductBlockOneForTest

    SUBSCRIPTION_MODEL_REGISTRY["TestProductOne"] = ProductTypeOneForTest

    test_model = ProductTypeOneForTestInactive.from_product_id(product_id=test_product_one, customer_id=str(uuid4()))
    test_model.block = ProductBlockOneForTestInactive.new(subscription_id=test_model.subscription_id)

    product_block, product_sub_block = test_product_block_one_db

    # Check metadata
    with pytest.raises(
        ValueError, match=r"Cannot create instance of abstract class. Use one of {'ProductBlockOneForTest'}"
    ):
        AbstractProductBlockOneForTestInactive.new(subscription_id=test_model.subscription_id)
    assert AbstractProductBlockOneForTestInactive.name is None
    assert not hasattr(AbstractProductBlockOneForTestInactive, "description")
    assert not hasattr(AbstractProductBlockOneForTestInactive, "product_block_id")
    assert not hasattr(AbstractProductBlockOneForTestInactive, "tag")
    assert ProductBlockOneForTestInactive.name == "ProductBlockOneForTest"
    assert ProductBlockOneForTestInactive.description == "Test Block"
    assert ProductBlockOneForTestInactive.product_block_id == product_block.product_block_id
    assert ProductBlockOneForTestInactive.tag == "TEST"

    test_model.save()
    db.session.commit()

    test_model = AbstractProductTypeOneForTestInactive.from_subscription(test_model.subscription_id)
    assert isinstance(test_model, ProductTypeOneForTestInactive)
    assert isinstance(test_model.block, ProductBlockOneForTestInactive)

    test_model = ProductTypeOneForTestInactive.from_subscription(test_model.subscription_id)
    assert isinstance(test_model.block, ProductBlockOneForTestInactive)

    test_model.block.int_field = 1
    test_model.block.str_field = "bla"
    test_model.block.list_field = [1]

    test_model = SubscriptionModel.from_other_lifecycle(test_model, SubscriptionLifecycle.ACTIVE)
    test_model.save()
    assert isinstance(test_model.block, ProductBlockOneForTest)

    test_model = AbstractProductTypeOneForTest.from_subscription(test_model.subscription_id)
    assert isinstance(test_model.block, ProductBlockOneForTest)

    test_model = ProductTypeOneForTest.from_subscription(test_model.subscription_id)
    assert isinstance(test_model.block, ProductBlockOneForTest)

    block = AbstractProductBlockOneForTest.from_db(test_model.block.subscription_instance_id)
    assert block.dict() == test_model.block.dict()
    assert isinstance(block, ProductBlockOneForTest)

    block = ProductBlockOneForTest.from_db(test_model.block.subscription_instance_id)
    assert block.dict() == test_model.block.dict()
    assert isinstance(block, ProductBlockOneForTest)


def test_subscription_instance_list():
    T = TypeVar("T")

    class Max2List(SubscriptionInstanceList[T]):
        max_items = 2

    # Also check generic subclass
    class ConList(Max2List[T]):
        min_items = 1

    class Model(BaseModel):
        list_field: ConList[int]

    with pytest.raises(ValidationError):
        Model(list_field=["a"])

    Model(list_field=[1])


def test_is_constrained_list_type():
    class ListType(ConstrainedList):
        min_items = 1

    assert _is_constrained_list_type(ListType) is True
    assert _is_constrained_list_type(SubscriptionInstanceList[int]) is True
    assert _is_constrained_list_type(Optional[int]) is False
    assert _is_constrained_list_type(List[int]) is False


def test_diff_in_db_empty(test_product_one, test_product_type_one):
    ProductTypeOneForTestInactive, _, _ = test_product_type_one

    assert ProductTypeOneForTestInactive.diff_product_in_database(test_product_one) == {}


@pytest.mark.skipif(
    sys.version_info < (3, 10), reason="Python 3.10 changed behavior of object.__annotations__ on empty derived class"
)
def test_diff_in_db_when_no_fields(test_product_one, test_product_type_one):
    # As of Python 3.10, __annotations__ of an empty derived class no longer contain inherited fields
    # Skipping this test for <3.10 because we have no usecase for empty derived classes + it's very difficult to fix this

    class Wrong(SubscriptionModel):
        pass

    assert (
        Wrong.diff_product_in_database(test_product_one)
        == {
            "TestProductOne": {
                "missing_fixed_inputs_in_model": {"test_fixed_input"},
                "missing_product_blocks_in_model": {"ProductBlockOneForTest"},
            }
        }
        != {
            "TestProductOne": {
                "missing_in_depends_on_blocks": {
                    "ProductBlockOneForTest": {"missing_product_blocks_in_db": {"SubBlockOneForTest"}}
                },
                "missing_product_blocks_in_model": {"SubBlockOneForTest"},
            }
        }
    )


def test_diff_in_db_missing_in_db(test_product_type_one):
    ProductTypeOneForTestInactive, _, _ = test_product_type_one

    product = ProductTable(
        name="TestProductEmpty", description="Test ProductTable", product_type="Test", tag="TEST", status="active"
    )

    db.session.add(product)
    db.session.commit()

    assert ProductTypeOneForTestInactive.diff_product_in_database(product.product_id) == {
        "TestProductEmpty": {
            "missing_fixed_inputs_in_db": {"test_fixed_input"},
            "missing_in_depends_on_blocks": {
                "ProductBlockOneForTest": {
                    "missing_product_blocks_in_db": {"SubBlockOneForTest"},
                    "missing_resource_types_in_db": {"int_field", "list_field", "str_field"},
                },
                "SubBlockOneForTest": {"missing_resource_types_in_db": {"int_field", "str_field"}},
            },
            "missing_product_blocks_in_db": {"ProductBlockOneForTest"},
        }
    }


def test_from_other_lifecycle_abstract(test_product_one):
    class AbstractProductBlockOneForTestInactive(ProductBlockModel, product_block_name="ProductBlockOneForTest"):
        str_field: Optional[str] = None
        list_field: List[int] = Field(default_factory=list)

    class AbstractProductBlockOneForTestProvisioning(
        AbstractProductBlockOneForTestInactive, lifecycle=[SubscriptionLifecycle.PROVISIONING]
    ):
        str_field: Optional[str] = None
        list_field: List[int]

    class AbstractProductBlockOneForTest(
        AbstractProductBlockOneForTestProvisioning, lifecycle=[SubscriptionLifecycle.ACTIVE]
    ):
        str_field: str
        list_field: List[int]

    class ProductBlockOneForTestInactive(
        AbstractProductBlockOneForTestInactive, product_block_name="ProductBlockOneForTest"
    ):
        str_field: Optional[str] = None
        list_field: List[int] = Field(default_factory=list)
        int_field: Optional[int] = None

    class ProductBlockOneForTestProvisioning(
        ProductBlockOneForTestInactive,
        AbstractProductBlockOneForTestProvisioning,
        lifecycle=[SubscriptionLifecycle.PROVISIONING],
    ):
        str_field: Optional[str] = None
        list_field: List[int]
        int_field: int

    class ProductBlockOneForTest(
        ProductBlockOneForTestProvisioning, AbstractProductBlockOneForTest, lifecycle=[SubscriptionLifecycle.ACTIVE]
    ):
        str_field: str
        list_field: List[int]
        int_field: int

    block = ProductBlockOneForTestInactive.new(subscription_id=uuid4())
    assert isinstance(block, ProductBlockOneForTestInactive)

    block.int_field = 1
    block.str_field = "bla"
    block.list_field = [1]

    active_block = ProductBlockOneForTest._from_other_lifecycle(block, SubscriptionLifecycle.ACTIVE, uuid4())

    assert isinstance(active_block, AbstractProductBlockOneForTest)
    assert isinstance(active_block, ProductBlockOneForTest)
    assert active_block.db_model == block.db_model


def test_from_other_lifecycle_sub(test_product_one, test_product_block_one, test_product_sub_block_one):
    SubBlockOneForTestInactive, _, SubBlockOneForTest = test_product_sub_block_one
    ProductBlockOneForTestInactive, _, ProductBlockOneForTest = test_product_block_one
    subscription_id = uuid4()

    block = ProductBlockOneForTestInactive.new(
        subscription_id=subscription_id, int_field=1, str_field="bla", list_field=[1]
    )
    block.sub_block = SubBlockOneForTestInactive.new(subscription_id=subscription_id, int_field=1, str_field="bla")
    block.sub_block_2 = SubBlockOneForTestInactive.new(subscription_id=subscription_id, int_field=1, str_field="bla")
    block.sub_block_list = [
        SubBlockOneForTestInactive.new(subscription_id=subscription_id, int_field=1, str_field="bla")
    ]

    assert isinstance(block, ProductBlockOneForTestInactive)

    active_block = ProductBlockOneForTest._from_other_lifecycle(block, SubscriptionLifecycle.ACTIVE, uuid4())

    assert isinstance(active_block, ProductBlockOneForTest)
    assert isinstance(active_block.sub_block, SubBlockOneForTest)
    assert isinstance(active_block.sub_block_2, SubBlockOneForTest)
    assert isinstance(active_block.sub_block_list[0], SubBlockOneForTest)
    assert active_block.db_model == block.db_model
    assert active_block.sub_block.db_model == block.sub_block.db_model
    assert active_block.sub_block_2.db_model == block.sub_block_2.db_model
    assert active_block.sub_block_list[0].db_model == block.sub_block_list[0].db_model


def test_serializable_property():
    class DerivedDomainModel(DomainModel):
        @serializable_property  # type: ignore
        def double_int_field(self) -> int:
            # This property is serialized
            return 2 * self.int_field

        @property
        def triple_int_field(self) -> int:
            # This property is not serialized
            return 3 * self.int_field

        int_field: int

    block = DerivedDomainModel(int_field=13)

    assert block.dict() == {"int_field": 13, "double_int_field": 26}


def test_inherited_serializable_property():
    class ProvisioningDomainModel(DomainModel):
        @serializable_property  # type: ignore
        def double_int_field(self) -> int:
            return 2 * self.int_field

        @serializable_property  # type: ignore
        def triple_int_field(self) -> int:
            return 3 * self.int_field

        int_field: int

    class ActiveDomainModel(ProvisioningDomainModel):
        @serializable_property  # type: ignore
        def triple_int_field(self) -> int:
            # override the base property
            return 30 * self.int_field

    block = ActiveDomainModel(int_field=1)

    assert block.dict() == {"int_field": 1, "double_int_field": 2, "triple_int_field": 30}


def test_subscription_save_list_with_zero_values(
    test_product_type_one, test_product_sub_block_one, product_one_subscription_1
):
    _, _, ProductTypeOneForTest = test_product_type_one

    subscription = ProductTypeOneForTest.from_subscription(product_one_subscription_1)
    subscription.block.list_field = [10, 0, 20, 30, 40, 0, 0]
    subscription.save()
    assert subscription.block.list_field == [10, 0, 20, 30, 40, 0, 0]

    subscription = ProductTypeOneForTest.from_subscription(product_one_subscription_1)
    assert subscription.block.list_field == [10, 0, 20, 30, 40, 0, 0]


def test_subscription_save_bool_list_with_false_values(
    test_product_type_one, test_product_sub_block_one, product_one_subscription_1
):
    _, _, ProductTypeOneForTest = test_product_type_one

    subscription = ProductTypeOneForTest.from_subscription(product_one_subscription_1)
    subscription.block.list_field = [True, False, True, False]
    subscription.save()
    assert subscription.block.list_field == [True, False, True, False]

    subscription = ProductTypeOneForTest.from_subscription(product_one_subscription_1)
    assert subscription.block.list_field == [True, False, True, False]



"""Add subscription metadata workflow.

Revision ID: b1970225392d
Revises: e05bb1967eff
Create Date: 2023-05-25 09:22:46.491454

"""

import sqlalchemy as sa
from alembic import op
from sqlalchemy.dialects import postgresql
from sqlalchemy_utils.types.uuid import UUIDType

# revision identifiers, used by Alembic.
revision = "b1970225392d"
down_revision = "e05bb1967eff"
branch_labels = None
depends_on = None

METADATA_TABLE_NAME = "subscription_metadata"


def upgrade() -> None:
    op.create_table(
        METADATA_TABLE_NAME,
        sa.Column(
            "subscription_id",
            UUIDType(),
            nullable=False,
            index=True,
        ),
        sa.Column("metadata", postgresql.JSONB(astext_type=sa.Text()), nullable=False),  # type: ignore
        sa.ForeignKeyConstraint(["subscription_id"], ["subscriptions.subscription_id"], ondelete="CASCADE"),
    )


def downgrade() -> None:
    conn = op.get_bind()
    conn.execute(sa.text(f"DROP TABLE IF EXISTS {METADATA_TABLE_NAME}"))



# Copyright 2019-2023 SURF.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from http import HTTPStatus
from typing import Callable
from uuid import UUID

import structlog
from sqlalchemy.inspection import inspect

from orchestrator.api.error_handling import raise_status
from orchestrator.db import ProcessSubscriptionTable, ProcessTable, ProductTable, SubscriptionTable, db
from orchestrator.db.database import SearchQuery
from orchestrator.db.filters.filters import generic_filter
from orchestrator.db.filters.generic_filters import (
    generic_bool_filter,
    generic_is_like_filter,
    generic_values_in_column_filter,
)
from orchestrator.utils.helpers import to_camel

logger = structlog.get_logger(__name__)


def product_filter(query: SearchQuery, value: str) -> SearchQuery:
    process_subscriptions = (
        db.session.query(ProcessSubscriptionTable)
        .join(SubscriptionTable)
        .join(ProductTable)
        .filter(ProductTable.name.ilike("%" + value + "%"))
        .subquery()
    )
    return query.filter(ProcessTable.process_id == process_subscriptions.c.pid)


def tag_filter(query: SearchQuery, value: str) -> SearchQuery:
    tags = value.split("-")
    process_subscriptions = (
        db.session.query(ProcessSubscriptionTable)
        .join(SubscriptionTable)
        .join(ProductTable)
        .filter(ProductTable.tag.in_(tags))
        .subquery()
    )
    return query.filter(ProcessTable.process_id == process_subscriptions.c.pid)


def subscriptions_filter(query: SearchQuery, value: str) -> SearchQuery:
    process_subscriptions = (
        db.session.query(ProcessSubscriptionTable)
        .join(SubscriptionTable)
        .filter(SubscriptionTable.description.ilike("%" + value + "%"))
        .subquery()
    )
    return query.filter(ProcessTable.process_id == process_subscriptions.c.pid)


def subscription_id_filter(query: SearchQuery, value: str) -> SearchQuery:
    process_subscriptions = db.session.query(ProcessSubscriptionTable).join(SubscriptionTable)
    process_subscriptions = generic_is_like_filter(SubscriptionTable.subscription_id)(process_subscriptions, value)
    process_subscriptions = process_subscriptions.subquery()
    return query.filter(ProcessTable.process_id == process_subscriptions.c.pid)


def target_filter(query: SearchQuery, value: str) -> SearchQuery:
    targets = value.split("-")
    process_subscriptions = (
        db.session.query(ProcessSubscriptionTable)
        .filter(ProcessSubscriptionTable.workflow_target.in_(targets))
        .subquery()
    )
    return query.filter(ProcessTable.process_id == process_subscriptions.c.pid)


def organisation_filter(query: SearchQuery, value: str) -> SearchQuery:
    try:
        value_as_uuid = UUID(value)
    except (ValueError, AttributeError):
        msg = f"Not a valid organisation, must be a UUID: '{value}'"
        logger.debug(msg)
        raise_status(HTTPStatus.BAD_REQUEST, msg)

    process_subscriptions = (
        db.session.query(ProcessSubscriptionTable)
        .join(SubscriptionTable)
        .filter(SubscriptionTable.customer_id == value_as_uuid)
        .subquery()
    )
    return query.filter(ProcessTable.process_id == process_subscriptions.c.pid)


BASE_CAMEL = {to_camel(key): generic_is_like_filter(value) for key, value in inspect(ProcessTable).columns.items()}
BASE_SNAKE = {key: generic_is_like_filter(value) for key, value in inspect(ProcessTable).columns.items()}

PROCESS_FILTER_FUNCTIONS_BY_COLUMN: dict[str, Callable[[SearchQuery, str], SearchQuery]] = (
    BASE_CAMEL
    | BASE_SNAKE
    | {
        "isTask": generic_bool_filter(ProcessTable.is_task),
        "is_task": generic_bool_filter(ProcessTable.is_task),
        "assignee": generic_values_in_column_filter(ProcessTable.assignee),
        "lastStatus": generic_values_in_column_filter(ProcessTable.last_status),
        "product": product_filter,
        "productTag": tag_filter,
        "subscriptions": subscriptions_filter,
        "subscriptionId": subscription_id_filter,
        "subscription_id": subscription_id_filter,
        "target": target_filter,
        "organisation": organisation_filter,
        "istask": generic_bool_filter(ProcessTable.is_task),  # TODO: will be removed in 1.4
        "status": generic_values_in_column_filter(ProcessTable.last_status),  # TODO: will be removed in 1.4
        "last_status": generic_values_in_column_filter(ProcessTable.last_status),  # TODO: will be removed in 1.4
        "tag": tag_filter,  # TODO: will be removed in 1.4
        "creator": generic_is_like_filter(ProcessTable.created_by),  # TODO: will be removed in 1.4
    }
)

filter_processes = generic_filter(PROCESS_FILTER_FUNCTIONS_BY_COLUMN)



# Copyright 2022-2023 SURF.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from typing import Any, Type, Union

import structlog
from graphql import GraphQLField, GraphQLResolveInfo
from graphql.pyutils import Path
from strawberry.extensions import SchemaExtension
from strawberry.field import StrawberryField
from strawberry.types.types import TypeDefinition

from orchestrator.utils.helpers import to_camel

logger = structlog.get_logger()


DeprecatedPaths = dict[str, str]


def get_path_as_string(path: Path) -> str:
    if path.prev:
        return f"{get_path_as_string(path.prev)}/{path.key}"
    return f"{path.key}"


def get_root_path(path: Path) -> Path:
    while path.prev:
        path = path.prev
    return path


def get_field_deprecation(info: GraphQLResolveInfo) -> Union[str, None]:
    field = info.parent_type.fields.get(info.field_name)
    if isinstance(field, GraphQLField):
        return field.deprecation_reason
    return None


class DeprecationCheckerExtension(SchemaExtension):
    deprecated_queries: DeprecatedPaths = {}
    deprecated_mutations: DeprecatedPaths = {}

    def resolve(self, _next: Any, root: Any, info: GraphQLResolveInfo, *args: Any, **kwargs: Any) -> Any:
        pathstring = get_path_as_string(info.path)
        root_type = get_root_path(info.path).typename

        reason = None
        if root_type == "Query":
            reason = self.deprecated_queries.get(pathstring)
        elif root_type == "Mutation":
            reason = self.deprecated_mutations.get(pathstring)

        if reason:
            logger.warning("Use of deprecated path", type=root_type, path=pathstring, deprecation_reason=repr(reason))
        elif field_deprecation := get_field_deprecation(info):
            logger.warning(
                "Use of deprecated field",
                field=info.field_name,
                type=info.parent_type.name,
                path=pathstring,
                deprecation_reason=repr(field_deprecation),
            )

        return _next(root, info, *args, **kwargs)


def get_deprecated_paths(type_definition: TypeDefinition) -> DeprecatedPaths:
    """Find all deprecated paths in the given Strawberry type.

    Returns:
        mapping of deprecated paths to deprecation reasons
    """
    to_inspect: list[tuple[list[str], list[StrawberryField]]] = [([], type_definition.fields)]
    deprecated_paths = {}
    while to_inspect:
        path, fields = to_inspect.pop()
        for field in fields:
            field_path = path + [to_camel(field.name)]
            if field.deprecation_reason:
                deprecated_paths["/".join(field_path)] = field.deprecation_reason
            elif hasattr(field.type, "_type_definition") and field.type._type_definition.fields:
                to_inspect.append((field_path, field.type._type_definition.fields))
    return deprecated_paths


def make_deprecation_checker_extension(
    query: Union[Type, None] = None, mutation: Union[Type, None] = None
) -> type[DeprecationCheckerExtension]:
    def deprecations_for(_type: Union[Type, None]) -> DeprecatedPaths:
        type_def: Union[TypeDefinition, None] = getattr(_type, "_type_definition", None) if _type else None
        if not type_def:
            return {}
        return get_deprecated_paths(type_def) if type_def else {}

    deprecated_queries = deprecations_for(query)
    deprecated_mutations = deprecations_for(mutation)

    logger.debug("Deprecations", queries=",".join(deprecated_queries), mutations=",".join(deprecated_mutations))

    # Modify class in-place; a bit dirty, but not dirty enough to warrant some metaclass-fu?
    DeprecationCheckerExtension.deprecated_queries = deprecated_queries
    DeprecationCheckerExtension.deprecated_mutations = deprecated_mutations
    return DeprecationCheckerExtension



from orchestrator.utils.speed import speed_humanize


def test_speed_humanize_long():
    assert speed_humanize(0) == "0 Mbit/s"
    assert speed_humanize(1) == "1 Mbit/s"
    assert speed_humanize(1000) == "1 Gbit/s"
    assert speed_humanize(1000000) == "1 Tbit/s"
    assert speed_humanize(1000000000) == "1 Pbit/s"
    assert speed_humanize("1000") == "1 Gbit/s"
    assert speed_humanize("750") == "750 Mbit/s"
    assert speed_humanize("2 Gbit/s") == "2 Gbit/s"

    class Foo:
        def __str__(self) -> str:
            return "bar"

    assert speed_humanize(Foo()) == "bar"


def test_speed_humanize_short():
    assert speed_humanize(0, True) == "0M"
    assert speed_humanize(1, True) == "1M"
    assert speed_humanize(1000, True) == "1G"
    assert speed_humanize(1000000, True) == "1T"
    assert speed_humanize(1000000000, True) == "1P"
    assert speed_humanize("1000", True) == "1G"
    assert speed_humanize("750", True) == "750M"
    assert speed_humanize("2 Gbit/s", True) == "2 Gbit/s"


def test_speed_humanize_elan_bandwidth():
    res = speed_humanize("1000,750,1000")
    assert res == "1 Gbit/s,750 Mbit/s,1 Gbit/s"



from typing import Any, Callable, Iterable

from orchestrator.types import strEnum


def _esc_str(i: int) -> str:
    return f"\033[{i}m"


class COLOR(strEnum):
    RESET = _esc_str(0)
    BOLD = _esc_str(1)
    DIM = _esc_str(2)
    ITALIC = _esc_str(3)
    UNDERLINE = _esc_str(4)

    BLACK = _esc_str(30)
    RED = _esc_str(31)
    GREEN = _esc_str(32)
    YELLOW = _esc_str(33)
    BLUE = _esc_str(34)
    MAGENTA = _esc_str(35)
    CYAN = _esc_str(36)


def str_fmt(text: str, *, flags: Iterable[COLOR] = ()) -> str:
    return "".join(f for f in flags) + text + COLOR.RESET


def print_fmt(text: str, *, flags: Iterable[COLOR] = (), print_fn: Callable = print, **kwargs: Any) -> None:
    print_fn(str_fmt(text, flags=flags), **kwargs)


def noqa_print(s: str, **kwargs: Any) -> None:
    print(s, **kwargs)  # noqa: T201



# Copyright 2019-2020 SURF.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import TYPE_CHECKING, Any, Dict, List, Optional, Tuple, Type, TypeVar

import strawberry
import structlog

from orchestrator.types import SubscriptionLifecycle, strEnum

logger = structlog.get_logger(__name__)


@strawberry.enum
class ProductLifecycle(strEnum):
    ACTIVE = "active"
    PRE_PRODUCTION = "pre production"
    PHASE_OUT = "phase out"
    END_OF_LIFE = "end of life"


_sub_type_per_lifecycle: Dict[Tuple[Type, Optional[SubscriptionLifecycle]], Type] = {}


def register_specialized_type(cls: Type, lifecycle: Optional[List[SubscriptionLifecycle]] = None) -> None:
    if lifecycle:
        for lifecycle_state in lifecycle:
            _sub_type_per_lifecycle[(cls.__base_type__, lifecycle_state)] = cls
    else:
        _sub_type_per_lifecycle[(cls.__base_type__, None)] = cls


def lookup_specialized_type(block: Type, lifecycle: Optional[SubscriptionLifecycle]) -> Type:
    if not hasattr(block, "__base_type__"):
        raise ValueError("Cannot instantiate a class that has no __base_type__ attribute")

    specialized_block = _sub_type_per_lifecycle.get((block.__base_type__, lifecycle), None)
    if specialized_block is None:
        specialized_block = _sub_type_per_lifecycle.get((block.__base_type__, None), None)
    if specialized_block is None:
        specialized_block = block
    return specialized_block


def validate_lifecycle_status(
    product_block_field_name: str, product_block_field_type: Any, lifecycle_status: SubscriptionLifecycle
) -> None:
    specialized_type = lookup_specialized_type(product_block_field_type, lifecycle_status)
    if not issubclass(product_block_field_type, specialized_type):
        raise AssertionError(
            f"The lifecycle status of the type for the field: {product_block_field_name}, {specialized_type.__name__} (based on {product_block_field_type.__name__}) is not suitable for the lifecycle status ({lifecycle_status}) of this model"
        )


if TYPE_CHECKING:
    from orchestrator.domain.base import SubscriptionModel
else:
    SubscriptionModel = None
    DomainModel = None
T = TypeVar("T", bound=SubscriptionModel)



import json
from http import HTTPStatus
from typing import Union

from fastapi import Response

from test.unit_tests.helpers import assert_no_diff


def get_product_blocks_query(
    first: int = 10,
    after: int = 0,
    filter_by: Union[list[str], None] = None,
    sort_by: Union[list[dict[str, str]], None] = None,
) -> bytes:
    query = """
query ProductBlocksQuery($first: Int!, $after: Int!, $filterBy: [GraphqlFilter!], $sortBy: [GraphqlSort!]) {
  productBlocks(first: $first, after: $after, filterBy: $filterBy, sortBy: $sortBy) {
    page {
      name
      endDate
      description
      createdAt
      status
      tag
      resourceTypes {
        resourceType
        description
      }
      dependsOn {
        name
        description
      }
      inUseBy {
        name
        description
      }
    }
    pageInfo {
      endCursor
      hasNextPage
      hasPreviousPage
      startCursor
      totalItems
    }
  }
}
    """
    return json.dumps(
        {
            "operationName": "ProductBlocksQuery",
            "query": query,
            "variables": {
                "first": first,
                "after": after,
                "sortBy": sort_by if sort_by else [],
                "filterBy": filter_by if filter_by else [],
            },
        }
    ).encode("utf-8")


def test_product_blocks_query(test_client):
    data = get_product_blocks_query(first=2)
    response: Response = test_client.post("/api/graphql", content=data, headers={"Content-Type": "application/json"})

    assert HTTPStatus.OK == response.status_code
    result = response.json()
    product_blocks_data = result["data"]["productBlocks"]
    product_blocks = product_blocks_data["page"]
    pageinfo = product_blocks_data["pageInfo"]

    assert len(product_blocks) == 2

    assert pageinfo == {
        "hasPreviousPage": False,
        "hasNextPage": True,
        "startCursor": 0,
        "endCursor": 1,
        "totalItems": 6,
    }

    expected = [
        {
            "name": "PB_1",
            "endDate": None,
            "description": "Generic Product Block 1",
            "createdAt": "2023-05-24T00:00:00+00:00",
            "status": "ACTIVE",
            "tag": "PB1",
            "resourceTypes": [{"resourceType": "rt_1", "description": "Resource Type one"}],
            "dependsOn": [],
            "inUseBy": [],
        },
        {
            "name": "PB_2",
            "endDate": None,
            "description": "Generic Product Block 2",
            "createdAt": "2023-05-24T00:00:00+00:00",
            "status": "ACTIVE",
            "tag": "PB2",
            "resourceTypes": [
                {"resourceType": "rt_2", "description": "Resource Type two"},
                {"resourceType": "rt_3", "description": "Resource Type three"},
            ],
            "dependsOn": [],
            "inUseBy": [],
        },
    ]
    assert product_blocks == expected


def test_product_block_query_with_relations(test_client):
    data = get_product_blocks_query(filter_by={"field": "name", "value": "ForTest"})
    response: Response = test_client.post("/api/graphql", content=data, headers={"Content-Type": "application/json"})

    assert HTTPStatus.OK == response.status_code
    result = response.json()
    product_blocks_data = result["data"]["productBlocks"]
    product_blocks = product_blocks_data["page"]
    pageinfo = product_blocks_data["pageInfo"]

    assert len(product_blocks) == 3

    assert pageinfo == {
        "hasPreviousPage": False,
        "hasNextPage": False,
        "startCursor": 0,
        "endCursor": 2,
        "totalItems": 3,
    }

    expected = [
        {
            "name": "SubBlockOneForTest",
            "endDate": None,
            "description": "Test Sub Block One",
            "status": "ACTIVE",
            "tag": "TEST",
            "resourceTypes": [
                {"resourceType": "int_field", "description": ""},
                {"resourceType": "str_field", "description": ""},
            ],
            "dependsOn": [],
            "inUseBy": [{"name": "ProductBlockWithListUnionForTest", "description": "Test Union Sub Block"}],
        },
        {
            "name": "SubBlockTwoForTest",
            "endDate": None,
            "description": "Test Sub Block Two",
            "status": "ACTIVE",
            "tag": "TEST",
            "resourceTypes": [{"resourceType": "int_field_2", "description": ""}],
            "dependsOn": [],
            "inUseBy": [{"name": "ProductBlockWithListUnionForTest", "description": "Test Union Sub Block"}],
        },
        {
            "name": "ProductBlockWithListUnionForTest",
            "endDate": None,
            "description": "Test Union Sub Block",
            "status": "ACTIVE",
            "tag": "TEST",
            "resourceTypes": [
                {"resourceType": "int_field", "description": ""},
                {"resourceType": "str_field", "description": ""},
                {"resourceType": "list_field", "description": ""},
            ],
            "dependsOn": [
                {"name": "SubBlockOneForTest", "description": "Test Sub Block One"},
                {"name": "SubBlockTwoForTest", "description": "Test Sub Block Two"},
            ],
            "inUseBy": [],
        },
    ]

    exclude_paths = exclude_paths = [f"root[{i}]['createdAt']" for i in range(len(expected))]
    assert_no_diff(expected, product_blocks, exclude_paths=exclude_paths)


def test_product_blocks_has_previous_page(test_client):
    data = get_product_blocks_query(after=1)
    response: Response = test_client.post("/api/graphql", content=data, headers={"Content-Type": "application/json"})

    assert HTTPStatus.OK == response.status_code
    result = response.json()
    product_blocks_data = result["data"]["productBlocks"]
    product_blocks = product_blocks_data["page"]
    pageinfo = product_blocks_data["pageInfo"]

    assert pageinfo == {
        "hasNextPage": False,
        "hasPreviousPage": True,
        "startCursor": 1,
        "endCursor": 5,
        "totalItems": 6,
    }

    assert len(product_blocks) == 5
    assert product_blocks[0]["name"] == "PB_2"
    assert product_blocks[1]["name"] == "PB_3"


def test_product_blocks_filter_by_resource_types(test_client):
    data = get_product_blocks_query(
        filter_by=[{"field": "resource_types", "value": "rt_1"}],
        sort_by=[{"field": "name", "order": "ASC"}],
    )
    response: Response = test_client.post("/api/graphql", content=data, headers={"Content-Type": "application/json"})
    assert HTTPStatus.OK == response.status_code
    result = response.json()
    product_blocks_data = result["data"]["productBlocks"]
    product_blocks = product_blocks_data["page"]
    pageinfo = product_blocks_data["pageInfo"]

    assert pageinfo == {
        "endCursor": 0,
        "hasNextPage": False,
        "hasPreviousPage": False,
        "startCursor": 0,
        "totalItems": 1,
    }
    assert product_blocks[0]["name"] == "PB_1"


def test_product_blocks_filter_by_products(test_client):
    data = get_product_blocks_query(
        filter_by=[{"field": "products", "value": "Product 1-Product 3"}],
        sort_by=[{"field": "name", "order": "ASC"}],
    )
    response: Response = test_client.post("/api/graphql", content=data, headers={"Content-Type": "application/json"})
    assert HTTPStatus.OK == response.status_code
    result = response.json()
    product_blocks_data = result["data"]["productBlocks"]
    product_blocks = product_blocks_data["page"]
    pageinfo = product_blocks_data["pageInfo"]

    assert pageinfo == {
        "hasNextPage": False,
        "hasPreviousPage": False,
        "startCursor": 0,
        "endCursor": 1,
        "totalItems": 2,
    }
    assert product_blocks[0]["name"] == "PB_1"
    assert product_blocks[1]["name"] == "PB_2"


def test_product_blocks_sort_by_tag(test_client):
    data = get_product_blocks_query(sort_by=[{"field": "tag", "order": "DESC"}])
    response: Response = test_client.post("/api/graphql", content=data, headers={"Content-Type": "application/json"})
    assert HTTPStatus.OK == response.status_code
    result = response.json()

    product_blocks_data = result["data"]["productBlocks"]
    product_blocks = product_blocks_data["page"]
    pageinfo = product_blocks_data["pageInfo"]

    assert [p_block["tag"] for p_block in product_blocks] == ["TEST", "TEST", "TEST", "PB3", "PB2", "PB1"]
    assert pageinfo == {
        "hasNextPage": False,
        "hasPreviousPage": False,
        "startCursor": 0,
        "endCursor": 5,
        "totalItems": 6,
    }



# Copyright 2019-2020 SURF.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from more_itertools import flatten

from orchestrator.cli.generator.generator.helpers import root_product_block


def get_all_validations(config: dict) -> list:
    def to_validations(field: dict) -> list[dict]:
        def to_validation(validation: dict) -> dict:
            return {"validation": validation, "field": field}

        return [to_validation(validation) for validation in field.get("validations", [])]

    product_block = root_product_block(config)
    return list(flatten(to_validations(field) for field in product_block["fields"]))


def get_validation_imports(validations: list) -> list:
    def format_validator(validation: dict) -> str:
        return f'{validation["validation"]["id"]}_validator'

    return [format_validator(validation) for validation in validations]


def get_validations(config: dict) -> tuple[list, list]:
    validations = get_all_validations(config)
    validation_imports = get_validation_imports(validations)

    return validations, validation_imports


def get_validations_for_modify(config: dict) -> tuple[list, list]:
    def modifiable(validation: dict) -> bool:
        return validation["field"].get("modifiable", True)

    validations = [validation for validation in get_all_validations(config) if modifiable(validation)]
    validation_imports = get_validation_imports(validations)

    return validations, validation_imports



# Copyright 2019-2020 SURF.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import re
from collections.abc import Generator

from more_itertools import first, one

from orchestrator.cli.generator.generator.settings import product_generator_settings


def snake_to_camel(s: str) -> str:
    return "".join(x.title() for x in s.split("_"))


def camel_to_snake(s: str) -> str:
    name = re.sub("(.)([A-Z][a-z]+)", r"\1_\2", s)
    return re.sub("([a-z0-9])([A-Z])", r"\1_\2", name).lower()


def get_workflow(config: dict, workflow_name: str) -> dict:
    workflows = (workflow for workflow in config.get("workflows", []) if workflow["name"] == workflow_name)
    return first(workflows, {})


def get_variable(config: dict) -> str:
    return config.get("variable", camel_to_snake(config["name"]))


def get_product_block_variable(product_block: dict) -> str:
    return get_variable(product_block)


def get_product_file_name(config: dict) -> str:
    return get_variable(config)


def get_product_block_file_name(product_block: dict) -> str:
    return get_product_block_variable(product_block)


def root_product_block(config: dict) -> dict:
    product_blocks = config.get("product_blocks", [])
    # TODO: multiple product_blocks will need more logic, ok for now
    return one(product_blocks)


def insert_into_imports(content: list[str], new_import: str) -> list[str]:
    # Note: we may consider using a real Python parser here someday, but for now this is ok and formatting
    # gets done by isort and black.
    def produce() -> Generator:
        not_inserted_yet = True
        for line in content:
            if line.startswith("from ") and not_inserted_yet:
                yield new_import
                not_inserted_yet = False
            yield line

    return list(produce())


def path_to_module(path: str) -> str:
    return path.replace("/", ".")


product_types_module = path_to_module(product_generator_settings.PRODUCT_TYPES_PATH)


def get_product_import(product: dict, lifecycle: str = "") -> str:
    return f'from {product_types_module}.{product["variable"]} import {product["type"]}{lifecycle}\n'



from typing import TYPE_CHECKING, Annotated, List

import strawberry

from orchestrator.schemas.resource_type import ResourceTypeSchema

if TYPE_CHECKING:
    from orchestrator.graphql.schemas.product_block import ProductBlock


@strawberry.experimental.pydantic.type(model=ResourceTypeSchema, all_fields=True)
class ResourceType:
    @strawberry.field(description="Return all product blocks that make use of this resource type")  # type: ignore
    async def product_blocks(self) -> List[Annotated["ProductBlock", strawberry.lazy(".product_block")]]:
        from orchestrator.graphql.schemas.product_block import ProductBlock

        return [ProductBlock.from_pydantic(product_block) for product_block in self._original_model.product_blocks]  # type: ignore



# Copyright 2019-2020 SURF.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Module that implements user (e.g. miscellaneous) related API endpoints."""

from typing import Any, Dict

import structlog
from fastapi.param_functions import Body
from fastapi.routing import APIRouter

from orchestrator.utils.json import json_loads

logger = structlog.get_logger(__name__)


router = APIRouter()


@router.post("/error", response_model=Dict)
def log_error(data: Dict[Any, Any] = Body(...)) -> Dict:
    logger.error("Client reported an error", data=data, frontend_error=True)
    return {}


@router.post("/log/{user_name}", response_model=Dict)
def log_user_info(user_name: str, message: Dict = Body(...)) -> Dict:
    """Log frontend messages that are related to user actions.

    When the frontend finalizes the setup of a login session it will do a HTTP POST to this endpoint. The frontend
    will also post to this endpoint when it ends a user session.

    Args:
        user_name: the username (email) of the user involved
        message: A log message.

    Returns:
        {}

    """
    try:
        json_dict: Any = json_loads(str(message))
        _message = json_dict["message"]
    except Exception:
        _message = message
    logger.info("Client sent user info", message=_message, user_name=user_name)
    return {}



from typing import TYPE_CHECKING, Annotated, List

import strawberry

from orchestrator.schemas import WorkflowSchema

if TYPE_CHECKING:
    from orchestrator.graphql.schemas.product import ProductType


@strawberry.experimental.pydantic.type(model=WorkflowSchema, all_fields=True)
class Workflow:
    @strawberry.field(description="Return all products that can use this workflow")  # type: ignore
    async def products(self) -> List[Annotated["ProductType", strawberry.lazy(".product")]]:
        from orchestrator.graphql.schemas.product import ProductType

        return [ProductType.from_pydantic(product) for product in self._original_model.products]  # type: ignore



from typing import Callable

import structlog

from orchestrator.db import ProductBlockTable, ProductTable, ResourceTypeTable
from orchestrator.db.database import SearchQuery
from orchestrator.db.filters import generic_filter
from orchestrator.db.filters.generic_filters import (
    generic_is_like_filter,
    generic_range_filters,
    generic_values_in_column_filter,
)

logger = structlog.get_logger(__name__)


def products_filter(query: SearchQuery, value: str) -> SearchQuery:
    """Filter ProductBlocks by '-'-separated list of Product 'name' (column) values."""
    products = value.split("-")
    return query.filter(ProductBlockTable.products.any(ProductTable.name.in_(products)))


def resource_types_filter(query: SearchQuery, value: str) -> SearchQuery:
    """Filter ProductBlocks by '-'-separated list of Resource Type 'resource_type' (column) values."""
    resource_types = value.split("-")
    return query.filter(ProductBlockTable.resource_types.any(ResourceTypeTable.resource_type.in_(resource_types)))


created_at_range_filters = generic_range_filters(ProductBlockTable.created_at)
end_date_range_filters = generic_range_filters(ProductBlockTable.end_date)

PRODUCT_BLOCK_FILTER_FUNCTIONS_BY_COLUMN: dict[str, Callable[[SearchQuery, str], SearchQuery]] = (
    {
        "product_block_id": generic_is_like_filter(ProductBlockTable.product_block_id),
        "name": generic_is_like_filter(ProductBlockTable.name),
        "description": generic_is_like_filter(ProductBlockTable.description),
        "tag": generic_values_in_column_filter(ProductBlockTable.tag),
        "status": generic_values_in_column_filter(ProductBlockTable.status),
        "products": products_filter,
        "resource_types": resource_types_filter,
    }
    | created_at_range_filters
    | end_date_range_filters
)

filter_product_blocks = generic_filter(PRODUCT_BLOCK_FILTER_FUNCTIONS_BY_COLUMN)



# Copyright 2019-2020 SURF.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from asyncio import new_event_loop
from typing import Any, Dict, Optional, cast
from urllib.parse import urlparse
from uuid import UUID

from structlog import get_logger

from orchestrator.db import ProcessTable
from orchestrator.settings import AppSettings, app_settings
from orchestrator.types import BroadcastFunc
from orchestrator.utils.enrich_process import enrich_process
from orchestrator.websocket.websocket_manager import WebSocketManager
from orchestrator.workflow import ProcessStat, ProcessStatus

logger = get_logger(__name__)


broadcaster_type = urlparse(app_settings.WEBSOCKET_BROADCASTER_URL).scheme


class WS_CHANNELS:
    ALL_PROCESSES = "processes"
    ENGINE_SETTINGS = "engine-settings"


async def empty_fn(*args: tuple, **kwargs: Dict[str, Any]) -> None:
    return


class WrappedWebSocketManager:
    def __init__(self, wrappee: Optional[WebSocketManager] = None) -> None:
        self.wrapped_websocket_manager = wrappee

    def update(self, wrappee: WebSocketManager) -> None:
        self.wrapped_websocket_manager = wrappee
        if self.wrapped_websocket_manager.enabled:
            logger.warning(
                "WebSocketManager object configured, all methods referencing `websocket_manager` should work."
            )
        else:
            logger.warning("WebSocketManager object not configured, ENABLE_WEBSOCKETS is false.")

    def __getattr__(self, attr: str) -> Any:
        if not isinstance(self.wrapped_websocket_manager, WebSocketManager):
            if "_" in attr:
                logger.warning("No WebSocketManager configured, but attempting to access class methods")
                return None
            raise RuntimeWarning(
                "No WebSocketManager configured at this time. Please pass WebSocketManager configuration to OrchestratorCore base_settings"
            )
        if attr != "enabled" and not self.wrapped_websocket_manager.enabled:
            logger.warning("Websockets are disabled, unable to access class methods")
            return empty_fn

        return getattr(self.wrapped_websocket_manager, attr)


# You need to pass a modified AppSettings class to the OrchestratorCore class to init the WebSocketManager correctly
wrapped_websocket_manager = WrappedWebSocketManager()
websocket_manager = cast(WebSocketManager, wrapped_websocket_manager)


# The Global WebSocketManager is set after calling this function
def init_websocket_manager(settings: AppSettings) -> WebSocketManager:
    wrapped_websocket_manager.update(WebSocketManager(settings.ENABLE_WEBSOCKETS, settings.WEBSOCKET_BROADCASTER_URL))
    return websocket_manager


def create_process_websocket_data(process: ProcessTable, pStat: ProcessStat) -> Dict:
    return {"process": enrich_process(process, pStat)}


def is_process_active(p: Dict) -> bool:
    return p["status"] in [ProcessStatus.RUNNING, ProcessStatus.SUSPENDED, ProcessStatus.WAITING]


def send_process_data_to_websocket(
    process_id: UUID,
    data: Dict,
    broadcast_func: Optional[BroadcastFunc] = None,
) -> None:
    """Broadcast data of the current process to connected websocket clients."""
    if not websocket_manager.enabled:
        return

    if broadcast_func:
        logger.debug("Broadcast process data through broadcast_func", process_id=str(process_id))
        broadcast_func(process_id, data)
    else:
        logger.debug("Broadcast process data directly to websocket_manager", process_id=str(process_id))
        loop = new_event_loop()
        channels = [WS_CHANNELS.ALL_PROCESSES]
        loop.run_until_complete(websocket_manager.broadcast_data(channels, data))
        try:
            loop.close()
        except Exception:  # noqa: S110
            pass


async def empty_handler() -> None:
    return


__all__ = [
    "websocket_manager",
    "init_websocket_manager",
    "create_process_websocket_data",
    "is_process_active",
    "send_process_data_to_websocket",
    "WS_CHANNELS",
]



from typing import Callable

from more_itertools import flatten, one
from strawberry.types.nodes import FragmentSpread, InlineFragment, SelectedField, Selection

from orchestrator.graphql.types import OrchestratorInfo


def get_selections(selected_field: Selection) -> list[Selection]:
    def has_field_name(selection: Selection, field_name: str) -> bool:
        return isinstance(selection, SelectedField) and selection.name == field_name

    page_field = [selection for selection in selected_field.selections if has_field_name(selection, "page")]

    if not page_field:
        return selected_field.selections
    return one(page_field).selections


def is_query_detailed(basic_fields: tuple) -> Callable[[OrchestratorInfo], bool]:
    """Function wrapper that validates GraphQL queries against provided basic_fields.

    Args:
        basic_fields: tuple of the basic fields.

    Returns:
        Function that validates if a GraphQL query includes fields beyond the specified basic_fields.
    """

    def _is_model_info_detailed(info: OrchestratorInfo) -> bool:
        """Check if the query asks for detailed props."""

        def has_details(selection: Selection) -> bool:
            if isinstance(selection, InlineFragment):
                return any(has_details(selection) for selection in selection.selections)
            if isinstance(selection, FragmentSpread):
                return any(has_details(s) for s in selection.selections)
            return selection.name not in basic_fields

        fields = flatten(get_selections(field) for field in info.selected_fields)
        return any(has_details(selection) for selection in fields if selection)

    return _is_model_info_detailed



from typing import Callable

import structlog

from orchestrator.db import ProductBlockTable, ResourceTypeTable
from orchestrator.db.database import SearchQuery
from orchestrator.db.filters import generic_filter
from orchestrator.db.filters.generic_filters import generic_is_like_filter

logger = structlog.get_logger(__name__)


def product_blocks_filter(query: SearchQuery, value: str) -> SearchQuery:
    """Filter ResourceTypes by '-'-separated list of Product block 'name' (column) values."""
    product_blocks = value.split("-")
    return query.filter(ResourceTypeTable.product_blocks.any(ProductBlockTable.name.in_(product_blocks)))


RESOURCE_TYPE_FILTER_FUNCTIONS_BY_COLUMN: dict[str, Callable[[SearchQuery, str], SearchQuery]] = {
    "resourceTypeId": generic_is_like_filter(ResourceTypeTable.resource_type_id),
    "resourceType": generic_is_like_filter(ResourceTypeTable.resource_type),
    "description": generic_is_like_filter(ResourceTypeTable.description),
    "productBlocks": product_blocks_filter,
}

filter_resource_types = generic_filter(RESOURCE_TYPE_FILTER_FUNCTIONS_BY_COLUMN)



# Copyright 2019-2020 SURF.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from datetime import datetime
from typing import Any, Dict, List, Optional
from uuid import UUID

from pydantic import ConstrainedStr

from orchestrator.config.assignee import Assignee
from orchestrator.schemas.base import OrchestratorBaseModel
from orchestrator.schemas.subscription import SubscriptionSchema
from orchestrator.targets import Target
from orchestrator.workflow import ProcessStatus


class ProcessIdSchema(OrchestratorBaseModel):
    id: UUID


class ProcessForm(OrchestratorBaseModel):
    title: str
    type: str
    properties: Dict[str, Any]
    additionalProperties: bool  # noqa: N815
    required: List[str] = []
    definitions: Optional[Dict[str, Any]]


class ProcessBaseSchema(OrchestratorBaseModel):
    process_id: UUID
    workflow_name: str
    is_task: bool
    created_by: Optional[str]
    failed_reason: Optional[str]
    started_at: datetime
    last_status: ProcessStatus
    last_step: Optional[str]
    assignee: Assignee
    last_modified_at: datetime
    traceback: Optional[str]

    class Config:
        orm_mode = True


class ProcessStepSchema(OrchestratorBaseModel):
    step_id: Optional[UUID]
    name: str
    status: str
    created_by: Optional[str] = None
    executed: Optional[datetime]
    commit_hash: Optional[str] = None
    state: Optional[Dict[str, Any]]

    stepid: Optional[UUID]  # TODO: will be removed in 1.4


class ProcessSchema(ProcessBaseSchema):
    product_id: Optional[UUID]
    customer_id: Optional[str]
    workflow_target: Optional[Target]
    subscriptions: List[SubscriptionSchema]
    current_state: Optional[Dict[str, Any]]
    steps: Optional[List[ProcessStepSchema]]
    form: Optional[ProcessForm]


class ProcessDeprecationsSchema(ProcessSchema):
    id: Optional[UUID]  # TODO: will be removed in 1.4
    pid: Optional[UUID]  # TODO: will be removed in 1.4
    workflow: Optional[str]  # TODO: will be removed in 1.4
    status: Optional[ProcessStatus]  # TODO: will be removed in 1.4
    step: Optional[str]  # TODO: will be removed in 1.4
    started: Optional[datetime]  # TODO: will be removed in 1.4
    last_modified: Optional[datetime]  # TODO: will be removed in 1.4
    product: Optional[UUID]  # TODO: will be removed in 1.4
    customer: Optional[str]  # TODO: will be removed in 1.4


class ProcessSubscriptionBaseSchema(OrchestratorBaseModel):
    id: UUID
    process_id: UUID
    subscription_id: UUID
    workflow_target: Optional[Target]
    created_at: datetime

    pid: UUID  # TODO: will be removed in 1.4

    class Config:
        orm_mode = True


class ProcessSubscriptionSchema(ProcessSubscriptionBaseSchema):
    process: ProcessBaseSchema


class ProcessResumeAllSchema(OrchestratorBaseModel):
    count: int


class ProcessStatusCounts(OrchestratorBaseModel):
    process_counts: Dict[ProcessStatus, int]
    task_counts: Dict[ProcessStatus, int]


class Reporter(ConstrainedStr):
    max_length = 100



from typing import Callable

from sqlalchemy import Column
from sqlalchemy.inspection import inspect
from sqlalchemy.sql import expression

from orchestrator.db import ProcessSubscriptionTable, ProcessTable, ProductTable, SubscriptionTable
from orchestrator.db.database import SearchQuery
from orchestrator.db.sorting.sorting import SortOrder, generic_column_sort, generic_sort
from orchestrator.utils.helpers import to_camel

PROCESS_CAMEL_SORT = {to_camel(key): generic_column_sort(value) for key, value in inspect(ProcessTable).columns.items()}
PROCESS_SNAKE_SORT = {key: generic_column_sort(value) for key, value in inspect(ProcessTable).columns.items()}


def generic_process_product_sort(field: Column) -> Callable[[SearchQuery, SortOrder], SearchQuery]:
    def sort_function(query: SearchQuery, order: SortOrder) -> SearchQuery:
        joined_query = query.join(ProcessSubscriptionTable).join(SubscriptionTable).join(ProductTable)
        if order == SortOrder.DESC:
            return joined_query.order_by(expression.desc(field))
        return joined_query.order_by(expression.asc(field))

    return sort_function


PROCESS_PRODUCT_SORT = {
    to_camel(key if "product" in key else f"product_{key}"): generic_process_product_sort(value)
    for key, value in inspect(ProductTable).columns.items()
}


PROCESS_SORT_FUNCTIONS_BY_COLUMN = (
    PROCESS_PRODUCT_SORT
    | PROCESS_CAMEL_SORT
    | PROCESS_SNAKE_SORT
    | {
        "workflow": generic_column_sort(ProcessTable.workflow_name),  # TODO: deprecated, remove in 1.4
        "status": generic_column_sort(ProcessTable.last_status),  # TODO: deprecated, remove in 1.4
        "creator": generic_column_sort(ProcessTable.created_by),  # TODO: deprecated, remove in 1.4
        "started": generic_column_sort(ProcessTable.started_at),  # TODO: deprecated, remove in 1.4
        "modified": generic_column_sort(ProcessTable.last_modified_at),  # TODO: deprecated, remove in 1.4
    }
)

sort_processes = generic_sort(PROCESS_SORT_FUNCTIONS_BY_COLUMN)



# Copyright 2019-2023 SURF.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Any, Callable, Iterator, Protocol

from more_itertools import partition
from pydantic import BaseModel

from orchestrator.api.error_handling import ProblemDetailException
from orchestrator.db.database import SearchQuery


class CallableErrorHandler(Protocol):
    def __call__(self, message: str, **kwargs: Any) -> None:
        ...


class Filter(BaseModel):
    field: str
    value: str


QueryType = SearchQuery
ValidFilterFunctionsByColumnType = dict[str, Callable[[QueryType, str], QueryType]]


def generic_filters_validate(
    valid_filter_functions_by_column: ValidFilterFunctionsByColumnType,
) -> Callable[[list[Filter]], tuple[Iterator[Filter], Iterator[Filter]]]:
    def _validate_filters(filter_by: list[Filter]) -> tuple[Iterator[Filter], Iterator[Filter]]:
        def _is_valid_filter(item: Filter) -> bool:
            return item.field in valid_filter_functions_by_column and item.value is not None

        return partition(_is_valid_filter, filter_by)

    return _validate_filters


def generic_apply_filters(
    valid_filter_functions_by_column: ValidFilterFunctionsByColumnType,
) -> Callable[[QueryType, Iterator[Filter], CallableErrorHandler], QueryType]:
    def _apply_filters(
        query: QueryType, filter_by: Iterator[Filter], handle_filter_error: CallableErrorHandler
    ) -> QueryType:
        for item in filter_by:
            field = item.field
            filter_fn = valid_filter_functions_by_column[field]
            try:
                query = filter_fn(query, item.value)
            except ProblemDetailException as exception:
                handle_filter_error(
                    exception.detail,
                    field=field,
                    value=item.value,
                )
            except ValueError as exception:
                handle_filter_error(
                    str(exception),
                    field=field,
                    value=item.value,
                )
        return query

    return _apply_filters


def generic_filter(
    valid_filter_functions_by_column: ValidFilterFunctionsByColumnType,
) -> Callable[[QueryType, list[Filter], CallableErrorHandler], QueryType]:
    valid_filter_functions_by_column_KEYS = list(valid_filter_functions_by_column.keys())
    _validate_filters = generic_filters_validate(valid_filter_functions_by_column)
    _apply_filters = generic_apply_filters(valid_filter_functions_by_column)

    def _filter(
        query: QueryType,
        filter_by: list[Filter],
        handle_filter_error: CallableErrorHandler,
    ) -> QueryType:
        invalid_filter_items, valid_filter_items = _validate_filters(filter_by)
        if invalid_list := [item.dict() for item in invalid_filter_items]:
            handle_filter_error(
                "Invalid filter arguments",
                invalid_filters=invalid_list,
                valid_filter_keys=valid_filter_functions_by_column_KEYS,
            )

        return _apply_filters(query, valid_filter_items, handle_filter_error)

    return _filter



from typing import Union

import structlog

from orchestrator.db.filters import Filter
from orchestrator.db.filters.product_block import filter_product_blocks
from orchestrator.db.models import ProductBlockTable
from orchestrator.db.range.range import apply_range_to_query
from orchestrator.db.sorting.product_block import sort_product_blocks
from orchestrator.db.sorting.sorting import Sort
from orchestrator.graphql.pagination import Connection
from orchestrator.graphql.schemas.product_block import ProductBlock
from orchestrator.graphql.types import GraphqlFilter, GraphqlSort, OrchestratorInfo
from orchestrator.graphql.utils.create_resolver_error_handler import create_resolver_error_handler
from orchestrator.graphql.utils.to_graphql_result_page import to_graphql_result_page

logger = structlog.get_logger(__name__)


async def resolve_product_blocks(
    info: OrchestratorInfo,
    filter_by: Union[list[GraphqlFilter], None] = None,
    sort_by: Union[list[GraphqlSort], None] = None,
    first: int = 10,
    after: int = 0,
) -> Connection[ProductBlock]:
    _error_handler = create_resolver_error_handler(info)

    pydantic_filter_by: list[Filter] = [item.to_pydantic() for item in filter_by] if filter_by else []
    pydantic_sort_by: list[Sort] = [item.to_pydantic() for item in sort_by] if sort_by else []
    logger.debug(
        "resolve_product_blocks() called", range=[after, after + first], sort=sort_by, filter=pydantic_filter_by
    )

    query = filter_product_blocks(ProductBlockTable.query, pydantic_filter_by, _error_handler)
    query = sort_product_blocks(query, pydantic_sort_by, _error_handler)
    total = query.count()
    query = apply_range_to_query(query, after, first)

    product_blocks = query.all()
    graphql_product_blocks = [ProductBlock.from_pydantic(p) for p in product_blocks]
    return to_graphql_result_page(graphql_product_blocks, first, after, total)



from typing import Annotated

import strawberry

from orchestrator.graphql.schemas.resource_type import ResourceType
from orchestrator.schemas.product_block import ProductBlockSchema


@strawberry.experimental.pydantic.type(model=ProductBlockSchema)
class ProductBlock:
    product_block_id: strawberry.auto
    name: strawberry.auto
    description: strawberry.auto
    tag: strawberry.auto
    status: strawberry.auto
    created_at: strawberry.auto
    end_date: strawberry.auto
    resource_types: list[ResourceType]

    @strawberry.field(description="Return all product blocks that this product block depends on")  # type: ignore
    async def depends_on(self) -> list[Annotated["ProductBlock", strawberry.lazy(".product_block")]]:
        return [ProductBlock.from_pydantic(product_block) for product_block in self._original_model.depends_on]  # type: ignore

    @strawberry.field(description="Return all product blocks that uses this product block")  # type: ignore
    async def in_use_by(self) -> list[Annotated["ProductBlock", strawberry.lazy(".product_block")]]:
        return [ProductBlock.from_pydantic(product_block) for product_block in self._original_model.in_use_by]  # type: ignore



"""Add throttling to refresh_subscriptions_view trigger.

Revision ID: a09ac125ea73
Revises: b1970225392d
Create Date: 2023-06-28 15:33:36.248121

"""
from pathlib import Path

from alembic import op
from sqlalchemy import text

# revision identifiers, used by Alembic.
revision = "a09ac125ea73"
down_revision = "b1970225392d"
branch_labels = None
depends_on = None


def upgrade() -> None:
    conn = op.get_bind()

    revision_file_path = Path(__file__)
    with open(revision_file_path.with_suffix(".sql")) as f:
        conn.execute(text(f.read()))


def downgrade() -> None:
    pass



from http import HTTPStatus
from os import getenv
from uuid import uuid4

import pytest
from redis import Redis

from nwastdlib.url import URL
from orchestrator.api.helpers import product_block_paths
from orchestrator.db import (
    FixedInputTable,
    ProcessSubscriptionTable,
    ProcessTable,
    ProductBlockTable,
    ProductTable,
    ResourceTypeTable,
    SubscriptionInstanceTable,
    SubscriptionInstanceValueTable,
    SubscriptionTable,
    db,
)
from orchestrator.db.models import SubscriptionInstanceRelationTable
from orchestrator.domain.base import SubscriptionModel
from orchestrator.services.subscriptions import (
    RELATION_RESOURCE_TYPES,
    _generate_etag,
    build_extended_domain_model,
    unsync,
)
from orchestrator.settings import app_settings
from orchestrator.utils.json import json_dumps, json_loads
from orchestrator.utils.redis import to_redis
from orchestrator.workflow import ProcessStatus
from test.unit_tests.config import (
    IMS_CIRCUIT_ID,
    INTERNETPINNEN_PREFIX_SUBSCRIPTION_ID,
    IPAM_PREFIX_ID,
    PARENT_IP_PREFIX_SUBSCRIPTION_ID,
    PEER_GROUP_SUBSCRIPTION_ID,
    PORT_SUBSCRIPTION_ID,
)

SERVICE_SUBSCRIPTION_ID = str(uuid4())
PORT_A_SUBSCRIPTION_ID = str(uuid4())
PORT_A_SUBSCRIPTION_BLOCK_ID = str(uuid4())
PROVISIONING_PORT_A_SUBSCRIPTION_ID = str(uuid4())
SSP_SUBSCRIPTION_ID = str(uuid4())
SSP_SUBSCRIPTION_BLOCK_ID = str(uuid4())
IP_PREFIX_SUBSCRIPTION_ID = str(uuid4())
INVALID_SUBSCRIPTION_ID = str(uuid4())
INVALID_PORT_SUBSCRIPTION_ID = str(uuid4())

PRODUCT_ID = str(uuid4())
CUSTOMER_ID = str(uuid4())


@pytest.fixture
def seed():
    # These resource types are special
    resources = [
        ResourceTypeTable(resource_type=IMS_CIRCUIT_ID, description="Desc"),
        ResourceTypeTable(resource_type=PORT_SUBSCRIPTION_ID, description="Desc"),
    ]
    product_blocks = [
        ProductBlockTable(name="ProductBlockA", description="description a", status="active", resource_types=resources)
    ]
    fixed_inputs = [FixedInputTable(name="product_type", value="MSP100G")]
    # Wf needs to exist in code and needs to be a legacy wf with mapping

    lp_product = ProductTable(
        name="LightpathProduct",
        description="Service product that lives on ports",
        product_type="LightPath",
        tag="LightPath",  # This is special since in_use_by/depends_on subscription code handles specific tags
        status="active",
        product_blocks=product_blocks,
        fixed_inputs=fixed_inputs,
    )
    port_a_product = ProductTable(
        product_id=PRODUCT_ID,
        name="PortAProduct",
        description="Port A description",
        product_type="Port",
        tag="SP",  # This is special since in_use_by/depends_on subscription code handles specific tags
        status="active",
        product_blocks=product_blocks,
        fixed_inputs=fixed_inputs,
    )
    port_b_product = ProductTable(
        name="PortBProduct",
        description="Port B description",
        product_type="Port",
        tag="PORTB",
        status="active",
        product_blocks=product_blocks,
        fixed_inputs=fixed_inputs,
    )

    # Special resource type handled by get subscription by ipam prefix endpoint
    ip_prefix_resources = [ResourceTypeTable(resource_type=IPAM_PREFIX_ID, description="Prefix id")]
    ip_prefix_product_blocks = [
        ProductBlockTable(
            name="ProductBlockB", description="description b", status="active", resource_types=ip_prefix_resources
        )
    ]
    ip_prefix_product = ProductTable(
        name="IPPrefixProduct",
        description="ProductTable that is used by service product",
        product_type="IP_PREFIX",
        tag="IP_PREFIX",
        status="active",
        product_blocks=ip_prefix_product_blocks,
    )
    ip_prefix_subscription = SubscriptionTable(
        subscription_id=IP_PREFIX_SUBSCRIPTION_ID,
        description="desc",
        status="active",
        insync=True,
        product=ip_prefix_product,
        customer_id=CUSTOMER_ID,
        instances=[
            SubscriptionInstanceTable(
                product_block=ip_prefix_product_blocks[0],
                values=[SubscriptionInstanceValueTable(resource_type=ip_prefix_resources[0], value="26")],
            )
        ],
    )

    port_a_subscription = SubscriptionTable(
        subscription_id=PORT_A_SUBSCRIPTION_ID,
        description="desc",
        status="initial",
        insync=True,
        product=port_a_product,
        customer_id=CUSTOMER_ID,
        instances=[
            SubscriptionInstanceTable(
                product_block=product_blocks[0],
                values=[SubscriptionInstanceValueTable(resource_type=resources[0], value="54321")],
            )
        ],
    )

    provisioning_port_a_subscription = SubscriptionTable(
        subscription_id=PROVISIONING_PORT_A_SUBSCRIPTION_ID,
        description="desc",
        status="provisioning",
        insync=False,
        product=port_a_product,
        customer_id=CUSTOMER_ID,
        instances=[
            SubscriptionInstanceTable(
                product_block=product_blocks[0],
                values=[SubscriptionInstanceValueTable(resource_type=resources[0], value="12345")],
            )
        ],
    )

    ssp_subscription = SubscriptionTable(
        subscription_id=SSP_SUBSCRIPTION_ID,
        description="desc",
        status="active",
        insync=True,
        product=port_b_product,
        customer_id=CUSTOMER_ID,
        instances=[
            SubscriptionInstanceTable(
                product_block=product_blocks[0],
                values=[SubscriptionInstanceValueTable(resource_type=resources[0], value="54321")],
            )
        ],
    )

    lp_subscription_instance_values_ssp = [
        SubscriptionInstanceValueTable(resource_type=resources[0], value="54321"),
        SubscriptionInstanceValueTable(resource_type=resources[1], value=str(PORT_A_SUBSCRIPTION_ID)),
    ]
    lp_subscription_instance_ssp = SubscriptionInstanceTable(
        product_block=product_blocks[0], values=lp_subscription_instance_values_ssp
    )
    lp_subscription_instance_values_msp = [
        SubscriptionInstanceValueTable(resource_type=resources[0], value="54321"),
        SubscriptionInstanceValueTable(resource_type=resources[1], value=str(SSP_SUBSCRIPTION_ID)),
    ]
    lp_subscription_instance_msp = SubscriptionInstanceTable(
        product_block=product_blocks[0], values=lp_subscription_instance_values_msp
    )
    lp_subscription = SubscriptionTable(
        subscription_id=SERVICE_SUBSCRIPTION_ID,
        description="desc",
        status="active",
        insync=True,
        product=lp_product,
        customer_id=CUSTOMER_ID,
        instances=[lp_subscription_instance_ssp, lp_subscription_instance_msp],
    )

    invalid_subscription = SubscriptionTable(
        subscription_id=INVALID_SUBSCRIPTION_ID,
        description="desc",
        status="active",
        insync=True,
        product=port_a_product,
        customer_id=CUSTOMER_ID,
        instances=[],
    )

    invalid_tagged_product = ProductTable(
        product_id=str(uuid4()),
        name="INVALID_PRODUCT",
        description="invalid descr",
        product_type="Port",
        tag="NEWMSP",
        status="active",
        product_blocks=product_blocks,
        fixed_inputs=fixed_inputs,
    )

    invalid_tagged_subscription = SubscriptionTable(
        subscription_id=INVALID_PORT_SUBSCRIPTION_ID,
        description="desc",
        status="active",
        insync=False,
        product=invalid_tagged_product,
        customer_id=CUSTOMER_ID,
        instances=[
            SubscriptionInstanceTable(
                product_block=product_blocks[0],
                values=[SubscriptionInstanceValueTable(resource_type=resources[0], value="54321")],
            )
        ],
    )
    db.session.add(port_a_product)
    db.session.add(ip_prefix_product)
    db.session.add(ip_prefix_subscription)
    db.session.add(invalid_tagged_product)
    db.session.add(port_b_product)
    db.session.add(lp_product)
    db.session.add(port_a_subscription)
    db.session.add(provisioning_port_a_subscription)
    db.session.add(ssp_subscription)
    db.session.add(lp_subscription)
    db.session.add(invalid_subscription)
    db.session.add(invalid_tagged_subscription)
    db.session.commit()

    RELATION_RESOURCE_TYPES.extend(
        [
            PORT_SUBSCRIPTION_ID,
            IP_PREFIX_SUBSCRIPTION_ID,
            INTERNETPINNEN_PREFIX_SUBSCRIPTION_ID,
            PARENT_IP_PREFIX_SUBSCRIPTION_ID,
            PEER_GROUP_SUBSCRIPTION_ID,
        ]
    )

    return ip_prefix_product


# seeder to test direct relations using SubscriptionInstanceRelationTable instead of resource type.
@pytest.fixture
def seed_with_direct_relations():
    # These resource types are special
    resources = [
        ResourceTypeTable(resource_type=IMS_CIRCUIT_ID, description="Desc"),
    ]
    product_blocks = [
        ProductBlockTable(name="ProductBlockA", description="description a", status="active", resource_types=resources)
    ]
    fixed_inputs = [FixedInputTable(name="product_type", value="MSP100G")]
    # Wf needs to exist in code and needs to be a legacy wf with mapping

    lp_product = ProductTable(
        name="LightpathProduct",
        description="Service product that lives on ports",
        product_type="LightPath",
        tag="LightPath",  # This is special since in_use_by/depends_on subscription code handles specific tags
        status="active",
        product_blocks=product_blocks,
        fixed_inputs=fixed_inputs,
    )
    port_a_product = ProductTable(
        product_id=PRODUCT_ID,
        name="PortAProduct",
        description="Port A description",
        product_type="Port",
        tag="SP",  # This is special since in_use_by/depends_on subscription code handles specific tags
        status="active",
        product_blocks=product_blocks,
        fixed_inputs=fixed_inputs,
    )
    port_b_product = ProductTable(
        name="PortBProduct",
        description="Port B description",
        product_type="Port",
        tag="PORTB",
        status="active",
        product_blocks=product_blocks,
        fixed_inputs=fixed_inputs,
    )

    # Special resource type handled by get subscription by ipam prefix endpoint
    ip_prefix_resources = [ResourceTypeTable(resource_type=IPAM_PREFIX_ID, description="Prefix id")]
    ip_prefix_product_blocks = [
        ProductBlockTable(
            name="ProductBlockB", description="description b", status="active", resource_types=ip_prefix_resources
        )
    ]
    ip_prefix_product = ProductTable(
        name="IPPrefixProduct",
        description="ProductTable that is used by service product",
        product_type="IP_PREFIX",
        tag="IP_PREFIX",
        status="active",
        product_blocks=ip_prefix_product_blocks,
    )
    ip_prefix_subscription = SubscriptionTable(
        subscription_id=IP_PREFIX_SUBSCRIPTION_ID,
        description="desc",
        status="active",
        insync=True,
        product=ip_prefix_product,
        customer_id=CUSTOMER_ID,
        instances=[
            SubscriptionInstanceTable(
                product_block=ip_prefix_product_blocks[0],
                values=[SubscriptionInstanceValueTable(resource_type=ip_prefix_resources[0], value="26")],
            )
        ],
    )

    port_a_subscription = SubscriptionTable(
        subscription_id=PORT_A_SUBSCRIPTION_ID,
        description="desc",
        status="initial",
        insync=True,
        product=port_a_product,
        customer_id=CUSTOMER_ID,
        instances=[
            SubscriptionInstanceTable(
                subscription_instance_id=PORT_A_SUBSCRIPTION_BLOCK_ID,
                product_block=product_blocks[0],
                values=[SubscriptionInstanceValueTable(resource_type=resources[0], value="54321")],
            )
        ],
    )

    provisioning_port_a_subscription = SubscriptionTable(
        subscription_id=PROVISIONING_PORT_A_SUBSCRIPTION_ID,
        description="desc",
        status="provisioning",
        insync=False,
        product=port_a_product,
        customer_id=CUSTOMER_ID,
        instances=[
            SubscriptionInstanceTable(
                product_block=product_blocks[0],
                values=[SubscriptionInstanceValueTable(resource_type=resources[0], value="12345")],
            )
        ],
    )

    ssp_subscription = SubscriptionTable(
        subscription_id=SSP_SUBSCRIPTION_ID,
        description="desc",
        status="active",
        insync=True,
        product=port_b_product,
        customer_id=CUSTOMER_ID,
        instances=[
            SubscriptionInstanceTable(
                subscription_instance_id=SSP_SUBSCRIPTION_BLOCK_ID,
                product_block=product_blocks[0],
                values=[SubscriptionInstanceValueTable(resource_type=resources[0], value="54321")],
            )
        ],
    )

    lp_subscription_instance_ssp_id = str(uuid4())
    lp_subscription_instance_values_ssp = [
        SubscriptionInstanceValueTable(resource_type=resources[0], value="54321"),
    ]
    lp_subscription_instance_ssp = SubscriptionInstanceTable(
        subscription_instance_id=lp_subscription_instance_ssp_id,
        product_block=product_blocks[0],
        values=lp_subscription_instance_values_ssp,
    )
    lp_subscription_instance_ssp_depends_on = SubscriptionInstanceRelationTable(
        in_use_by_id=lp_subscription_instance_ssp_id,
        depends_on_id=PORT_A_SUBSCRIPTION_BLOCK_ID,
        order_id=0,
        domain_model_attr="service_port",
    )

    lp_subscription_instance_msp_id = str(uuid4())
    lp_subscription_instance_values_msp = [
        SubscriptionInstanceValueTable(resource_type=resources[0], value="54321"),
    ]
    lp_subscription_instance_msp = SubscriptionInstanceTable(
        subscription_instance_id=lp_subscription_instance_msp_id,
        product_block=product_blocks[0],
        values=lp_subscription_instance_values_msp,
    )
    lp_subscription_instance_msp_depends_on = SubscriptionInstanceRelationTable(
        in_use_by_id=lp_subscription_instance_msp_id,
        depends_on_id=SSP_SUBSCRIPTION_BLOCK_ID,
        order_id=0,
        domain_model_attr="service_port",
    )
    lp_subscription = SubscriptionTable(
        subscription_id=SERVICE_SUBSCRIPTION_ID,
        description="desc",
        status="active",
        insync=True,
        product=lp_product,
        customer_id=CUSTOMER_ID,
        instances=[lp_subscription_instance_ssp, lp_subscription_instance_msp],
    )

    invalid_subscription = SubscriptionTable(
        subscription_id=INVALID_SUBSCRIPTION_ID,
        description="desc",
        status="active",
        insync=True,
        product=port_a_product,
        customer_id=CUSTOMER_ID,
        instances=[],
    )

    invalid_tagged_product = ProductTable(
        product_id=str(uuid4()),
        name="INVALID_PRODUCT",
        description="invalid descr",
        product_type="Port",
        tag="NEWMSP",
        status="active",
        product_blocks=product_blocks,
        fixed_inputs=fixed_inputs,
    )

    invalid_tagged_subscription = SubscriptionTable(
        subscription_id=INVALID_PORT_SUBSCRIPTION_ID,
        description="desc",
        status="active",
        insync=False,
        product=invalid_tagged_product,
        customer_id=CUSTOMER_ID,
        instances=[
            SubscriptionInstanceTable(
                product_block=product_blocks[0],
                values=[SubscriptionInstanceValueTable(resource_type=resources[0], value="54321")],
            )
        ],
    )

    db.session.add(port_a_product)
    db.session.add(ip_prefix_product)
    db.session.add(ip_prefix_subscription)
    db.session.add(invalid_tagged_product)
    db.session.add(port_b_product)
    db.session.add(lp_product)
    db.session.add(port_a_subscription)
    db.session.add(provisioning_port_a_subscription)
    db.session.add(ssp_subscription)
    db.session.add(lp_subscription)
    db.session.add(invalid_subscription)
    db.session.add(invalid_tagged_subscription)
    db.session.commit()
    db.session.add(lp_subscription_instance_ssp_depends_on)
    db.session.add(lp_subscription_instance_msp_depends_on)
    db.session.commit()

    RELATION_RESOURCE_TYPES.extend(
        [
            PORT_SUBSCRIPTION_ID,
            IP_PREFIX_SUBSCRIPTION_ID,
            INTERNETPINNEN_PREFIX_SUBSCRIPTION_ID,
            PARENT_IP_PREFIX_SUBSCRIPTION_ID,
            PEER_GROUP_SUBSCRIPTION_ID,
        ]
    )

    return ip_prefix_product


def test_subscriptions_all(seed, test_client):
    product_fields = [
        "name",
        "created_at",
        "description",
        "end_date",
        "status",
        "product_type",
        "product_id",
        "tag",
    ]
    subscription_fields = [
        "customer_id",
        "description",
        "status",
        "end_date",
        "insync",
        "start_date",
        "subscription_id",
        "product",
        "name",
        "note",
        "customer_descriptions",
        "product_id",
        "tag",
    ]
    response = test_client.get("/api/subscriptions/all")

    assert response.status_code == HTTPStatus.OK
    assert len(response.json()) == 7
    for item in response.json():
        # Todo: determine if we want to let the test fail when extra fields are detected?
        product = item["product"]
        # test for extra fields in the response that are not listed in the YAML
        assert set(product.keys()) == set(product_fields)
        assert set(item.keys()) == set(subscription_fields)

        # Check if all listed YAML fields are also available in Response
        for field in product_fields:
            assert field in product
        for field in subscription_fields:
            assert field in item


def test_subscriptions_paginated(seed, test_client):
    product_fields = [
        "name",
        "created_at",
        "description",
        "end_date",
        "status",
        "product_type",
        "product_id",
        "tag",
    ]
    subscription_fields = [
        "customer_id",
        "description",
        "status",
        "end_date",
        "insync",
        "start_date",
        "subscription_id",
        "product",
        "name",
        "note",
        "customer_descriptions",
        "product_id",
        "tag",
    ]
    response = test_client.get("/api/subscriptions")

    assert response.status_code == HTTPStatus.OK
    assert len(response.json()) == 7

    response = test_client.get("/api/subscriptions/all")

    assert response.status_code == HTTPStatus.OK
    assert len(response.json()) == 7
    for item in response.json():
        product = item["product"]
        # test for extra fields in the response that are not listed in the YAML
        assert set(product.keys()) == set(product_fields)
        assert set(item.keys()) == set(subscription_fields)

        # Check if all listed YAML fields are also available in Response
        for field in product_fields:
            assert field in product
        for field in subscription_fields:
            assert field in item


def test_filtering_subscriptions(seed, test_client):
    response = test_client.get("/api/subscriptions?filter=status,active")
    assert response.status_code == HTTPStatus.OK
    assert len(response.json()) == 5

    response = test_client.get("/api/subscriptions?filter=insync,no")
    assert response.status_code == HTTPStatus.OK
    assert len(response.json()) == 2

    response = test_client.get("/api/subscriptions?filter=insync,Y")
    assert response.status_code == HTTPStatus.OK
    assert len(response.json()) == 5

    response = test_client.get("/api/subscriptions?filter=insync,no,status,provisioning")
    assert response.status_code == HTTPStatus.OK
    assert len(response.json()) == 1

    response = test_client.get("/api/subscriptions?filter=status_gt,active")
    assert response.status_code == HTTPStatus.OK
    assert len(response.json()) == 2

    response = test_client.get("/api/subscriptions?filter=status_gte,active")
    assert response.status_code == HTTPStatus.OK
    assert len(response.json()) == 7

    response = test_client.get("/api/subscriptions?filter=status_lt,initial")
    assert response.status_code == HTTPStatus.OK
    assert len(response.json()) == 5

    response = test_client.get("/api/subscriptions?filter=status_lte,initial")
    assert response.status_code == HTTPStatus.OK
    assert len(response.json()) == 6

    response = test_client.get("/api/subscriptions?filter=status,active,product,LightPathProduct")
    assert response.status_code == HTTPStatus.OK
    assert len(response.json()) == 1

    response = test_client.get("/api/subscriptions?filter=status,active,product,LightPathProduct-PortBProduct")
    assert response.status_code == HTTPStatus.OK
    assert len(response.json()) == 2


def test_sorting_subscriptions(seed, test_client):
    response = test_client.get("/api/subscriptions?sort=status,asc")
    assert response.status_code == HTTPStatus.OK
    assert len(response.json()) == 7
    assert response.json()[0]["status"] == "active"
    assert response.json()[6]["status"] == "provisioning"

    response = test_client.get("/api/subscriptions?sort=status,desc")
    assert response.status_code == HTTPStatus.OK
    assert len(response.json()) == 7
    assert response.json()[0]["status"] == "provisioning"
    assert response.json()[6]["status"] == "active"

    response = test_client.get("/api/subscriptions?sort=tag,asc")
    assert response.status_code == HTTPStatus.OK
    assert len(response.json()) == 7
    assert response.json()[0]["product"]["tag"] == "IP_PREFIX"
    assert response.json()[6]["product"]["tag"] == "SP"

    response = test_client.get("/api/subscriptions?sort=tag,desc")
    assert response.status_code == HTTPStatus.OK
    assert len(response.json()) == 7
    assert response.json()[0]["product"]["tag"] == "SP"
    assert response.json()[6]["product"]["tag"] == "IP_PREFIX"

    response = test_client.get("/api/subscriptions?sort=product,asc")
    assert response.status_code == HTTPStatus.OK
    assert len(response.json()) == 7
    assert response.json()[0]["product"]["name"] == "INVALID_PRODUCT"
    assert response.json()[6]["product"]["name"] == "PortBProduct"

    response = test_client.get("/api/subscriptions?sort=product,desc")
    assert response.status_code == HTTPStatus.OK
    assert len(response.json()) == 7
    assert response.json()[0]["product"]["name"] == "PortBProduct"
    assert response.json()[6]["product"]["name"] == "INVALID_PRODUCT"


def test_range_subscriptions(seed, test_client):
    response = test_client.get("/api/subscriptions?range=0,3")

    assert response.status_code == HTTPStatus.OK
    assert len(response.json()) == 3

    response = test_client.get("/api/subscriptions?sort=status,asc&range=5,5")
    assert response.status_code == HTTPStatus.BAD_REQUEST


def test_insync_404(seed, test_client):
    response = test_client.get(f"/api/subscriptions/workflows/{str(uuid4())}")
    assert response.status_code == HTTPStatus.NOT_FOUND


def test_insync_invalid_tagged(seed, test_client):
    response = test_client.get(f"/api/subscriptions/workflows/{INVALID_PORT_SUBSCRIPTION_ID}")
    assert response.status_code == HTTPStatus.OK
    assert response.json() == {
        "reason": "subscription.not_in_sync",
        "create": [],
        "modify": [],
        "terminate": [],
        "system": [],
    }


def test_in_use_by_subscriptions(seed, test_client):
    response = test_client.get(f"/api/subscriptions/in_use_by/{PORT_A_SUBSCRIPTION_ID}")
    in_use_by_subs = response.json()
    assert len(in_use_by_subs) == 1
    assert SERVICE_SUBSCRIPTION_ID == in_use_by_subs[0]["subscription_id"]


def test_subscriptions_for_in_used_by_ids(seed, test_client, caplog):
    instance_id = str(SubscriptionInstanceTable.query.first().subscription_instance_id)
    response = test_client.post("/api/subscriptions/subscriptions_for_in_used_by_ids", json=[instance_id])
    in_use_by_subs = response.json()
    assert in_use_by_subs[instance_id]
    assert in_use_by_subs[instance_id]["product"]["product_type"] == "IP_PREFIX"
    assert "Not all subscription_instance_id's could be resolved." not in caplog.text


def test_subscriptions_for_in_used_by_ids_with_wrong_instance_ids(seed, test_client, caplog):
    response = test_client.post(
        "/api/subscriptions/subscriptions_for_in_used_by_ids", json=["5373600d-c9ee-4ceb-96bd-1d3256baccec"]
    )
    in_use_by_subs = response.json()
    assert len(in_use_by_subs) == 0
    assert "Not all subscription_instance_id's could be resolved." in caplog.text
    assert "[UUID('5373600d-c9ee-4ceb-96bd-1d3256baccec')]" in caplog.text


def test_in_use_by_subscriptions_not_insync(seed, test_client):
    # ensure that the used subscription of the MSP is out of sync
    service = SubscriptionTable.query.get(SERVICE_SUBSCRIPTION_ID)
    service.insync = False
    db.session.commit()

    # test the api endpoint
    response = test_client.get(f"/api/subscriptions/workflows/{PORT_A_SUBSCRIPTION_ID}")
    assert response.status_code == HTTPStatus.OK
    insync_info = response.json()
    assert "reason" in insync_info
    assert len(insync_info["locked_relations"]) == 1
    assert insync_info["locked_relations"][0] == SERVICE_SUBSCRIPTION_ID


def test_in_use_by_subscriptions_insync(seed, test_client):
    response = test_client.get(f"/api/subscriptions/workflows/{PORT_A_SUBSCRIPTION_ID}")
    assert response.status_code == HTTPStatus.OK

    insync_info = response.json()
    assert "reason" not in insync_info
    assert "locked_relations" not in insync_info


def test_depends_on_subscriptions(seed, test_client):
    response = test_client.get(
        f"/api/subscriptions/depends_on/{SERVICE_SUBSCRIPTION_ID}?filter_statuses=initial,provisioning,active"
    )
    depends_on_subs = list(map(lambda sub: sub["subscription_id"], response.json()))
    assert len(depends_on_subs) == 2

    assert PORT_A_SUBSCRIPTION_ID in depends_on_subs
    assert SSP_SUBSCRIPTION_ID in depends_on_subs


def test_depends_on_subscriptions_bogus_status(seed, test_client):
    response = test_client.get(f"/api/subscriptions/depends_on/{SERVICE_SUBSCRIPTION_ID}?filter_statuses=NOT_VALID")
    assert response.status_code == 422
    assert response.json()["detail"] == "Status NOT_VALID, is not a valid `SubscriptionLifecycle`"


def test_depends_on_subscriptions_not_insync(seed, test_client):
    # ensure that the depends on subscription of the LP is out of sync
    msp = SubscriptionTable.query.with_for_update().get(PORT_A_SUBSCRIPTION_ID)
    msp.insync = False
    db.session.commit()

    # test the api endpoint
    response = test_client.get(f"/api/subscriptions/workflows/{SERVICE_SUBSCRIPTION_ID}")
    assert response.status_code == HTTPStatus.OK
    insync_info = response.json()
    assert "reason" in insync_info
    assert len(insync_info["locked_relations"]) == 1
    assert insync_info["locked_relations"][0] == PORT_A_SUBSCRIPTION_ID


def test_depends_on_subscriptions_insync(seed, test_client):
    response = test_client.get(f"/api/subscriptions/workflows/{SERVICE_SUBSCRIPTION_ID}")
    assert response.status_code == HTTPStatus.OK

    insync_info = response.json()
    assert "reason" not in insync_info
    assert "locked_relations" not in insync_info


def test_in_use_by_subscriptions_direct_relations(seed_with_direct_relations, test_client):
    response = test_client.get(f"/api/subscriptions/in_use_by/{PORT_A_SUBSCRIPTION_ID}")
    in_use_by_subs = response.json()
    assert len(in_use_by_subs) == 1
    assert SERVICE_SUBSCRIPTION_ID == in_use_by_subs[0]["subscription_id"]


def test_subscriptions_for_in_used_by_ids_direct_relations(seed_with_direct_relations, test_client, caplog):
    instance_id = str(SubscriptionInstanceTable.query.first().subscription_instance_id)
    response = test_client.post("/api/subscriptions/subscriptions_for_in_used_by_ids", json=[instance_id])
    depends_subs = response.json()
    assert depends_subs[instance_id]
    assert depends_subs[instance_id]["product"]["product_type"] == "IP_PREFIX"
    assert "Not all subscription_instance_id's could be resolved." not in caplog.text


def test_in_use_by_subscriptions_not_insync_direct_relations(seed_with_direct_relations, test_client):
    # ensure that the used subscription of the MSP is out of sync
    service = SubscriptionTable.query.get(SERVICE_SUBSCRIPTION_ID)
    service.insync = False
    db.session.commit()

    # test the api endpoint
    response = test_client.get(f"/api/subscriptions/workflows/{PORT_A_SUBSCRIPTION_ID}")
    assert response.status_code == HTTPStatus.OK
    insync_info = response.json()
    assert "reason" in insync_info
    assert len(insync_info["locked_relations"]) == 1
    assert insync_info["locked_relations"][0] == SERVICE_SUBSCRIPTION_ID


def test_in_use_by_subscriptions_insync_direct_relations(seed_with_direct_relations, test_client):
    response = test_client.get(f"/api/subscriptions/workflows/{PORT_A_SUBSCRIPTION_ID}")
    assert response.status_code == HTTPStatus.OK

    insync_info = response.json()
    assert "reason" not in insync_info
    assert "locked_relations" not in insync_info


def test_depends_on_subscriptions_direct_relations(seed_with_direct_relations, test_client):
    response = test_client.get(
        f"/api/subscriptions/depends_on/{SERVICE_SUBSCRIPTION_ID}?filter_statuses=initial,provisioning,active"
    )
    depends_on_subs = list(map(lambda sub: sub["subscription_id"], response.json()))
    assert len(depends_on_subs) == 2

    assert PORT_A_SUBSCRIPTION_ID in depends_on_subs
    assert SSP_SUBSCRIPTION_ID in depends_on_subs


def test_depends_on_subscriptions_not_insync_direct_relations(seed_with_direct_relations, test_client):
    # ensure that the depends on subscription of the LP is out of sync
    msp = SubscriptionTable.query.with_for_update().get(PORT_A_SUBSCRIPTION_ID)
    msp.insync = False
    db.session.commit()

    # test the api endpoint
    response = test_client.get(f"/api/subscriptions/workflows/{SERVICE_SUBSCRIPTION_ID}")
    assert response.status_code == HTTPStatus.OK
    insync_info = response.json()
    assert "reason" in insync_info
    assert len(insync_info["locked_relations"]) == 1
    assert insync_info["locked_relations"][0] == PORT_A_SUBSCRIPTION_ID


def test_depends_on_subscriptions_insync_direct_relations(seed_with_direct_relations, test_client):
    response = test_client.get(f"/api/subscriptions/workflows/{SERVICE_SUBSCRIPTION_ID}")
    assert response.status_code == HTTPStatus.OK

    insync_info = response.json()
    assert "reason" not in insync_info
    assert "locked_relations" not in insync_info


def test_delete_subscription(responses, seed, test_client):
    process_id = str(uuid4())
    db.session.add(
        ProcessTable(
            process_id=process_id, workflow_name="statisch_lichtpad_aanvragen", last_status=ProcessStatus.CREATED
        )
    )
    db.session.add(ProcessSubscriptionTable(process_id=process_id, subscription_id=PORT_A_SUBSCRIPTION_ID))
    db.session.commit()

    response = test_client.delete(f"/api/subscriptions/{PORT_A_SUBSCRIPTION_ID}")
    assert response.status_code == HTTPStatus.OK

    response = test_client.get("/api/processes")
    assert len(response.json()) == 0


def test_delete_subscription_404(responses, seed, test_client):
    sub_id = uuid4()

    response = test_client.delete(f"/api/subscriptions/{sub_id}")
    assert response.status_code == HTTPStatus.NOT_FOUND


def test_subscription_detail_with_domain_model(test_client, generic_subscription_1):
    # test with a subscription that has domain model and without
    response = test_client.get(URL("api/subscriptions/domain-model") / generic_subscription_1)
    assert response.status_code == HTTPStatus.OK
    # Check hierarchy
    assert response.json()["pb_1"]["rt_1"] == "Value1"


def test_subscription_detail_with_domain_model_does_not_exist(test_client, generic_subscription_1):
    # test with a subscription that has domain model and without
    response = test_client.get(URL("api/subscriptions/domain-model") / uuid4())
    assert response.status_code == HTTPStatus.NOT_FOUND


def test_subscription_detail_with_domain_model_etag(test_client, generic_subscription_1):
    # test with a subscription that has domain model and without
    response = test_client.get(URL("api/subscriptions/domain-model") / generic_subscription_1)
    assert response.status_code == HTTPStatus.OK
    subscription = SubscriptionModel.from_subscription(generic_subscription_1)
    extended_model = build_extended_domain_model(subscription)
    etag = _generate_etag(extended_model)
    assert etag == response.headers["ETag"]
    # Check hierarchy
    assert response.json()["pb_1"]["rt_1"] == "Value1"


def test_subscription_detail_with_domain_model_if_none_match(test_client, generic_subscription_1):
    # test with a subscription that has domain model and without
    subscription = SubscriptionModel.from_subscription(generic_subscription_1)
    extended_model = build_extended_domain_model(subscription)
    etag = _generate_etag(extended_model)
    response = test_client.get(
        URL("api/subscriptions/domain-model") / generic_subscription_1, headers={"If-None-Match": etag}
    )
    assert response.status_code == HTTPStatus.NOT_MODIFIED


@pytest.mark.skipif(
    not getenv("AIOCACHE_DISABLE", "0") == "0", reason="AIOCACHE must be enabled for this test to do anything"
)
def test_subscription_detail_with_domain_model_cache(test_client, generic_subscription_1):
    # test with a subscription that has domain model and without
    subscription = SubscriptionModel.from_subscription(generic_subscription_1)
    extended_model = build_extended_domain_model(subscription)
    etag = _generate_etag(extended_model)

    app_settings.CACHE_DOMAIN_MODELS = True

    to_redis(extended_model)

    response = test_client.get(URL("api/subscriptions/domain-model") / generic_subscription_1)

    cache = Redis.from_url(app_settings.CACHE_URI)
    result = cache.get(f"domain:{generic_subscription_1}")
    cached_model = json_dumps(json_loads(result))
    cached_etag = cache.get(f"domain:etag:{generic_subscription_1}")
    assert cached_model == json_dumps(extended_model)
    assert cached_etag.decode("utf-8") == etag

    assert response.status_code == HTTPStatus.OK
    assert response.json()["subscription_id"] == generic_subscription_1
    app_settings.CACHE_DOMAIN_MODELS = False
    cache.delete(f"domain:{generic_subscription_1}")


def test_subscription_detail_with_in_use_by_ids_filtered_self(test_client, product_one_subscription_1):
    response = test_client.get(URL("api/subscriptions/domain-model") / product_one_subscription_1)
    assert response.status_code == HTTPStatus.OK
    assert not response.json()["block"]["sub_block"]["in_use_by_ids"]


def test_subscription_detail_with_in_use_by_ids_not_filtered_self(test_client, product_one_subscription_1):
    response = test_client.get(
        URL("api/subscriptions/domain-model") / product_one_subscription_1 / "?filter_owner_relations=false"
    )
    assert response.status_code == HTTPStatus.OK
    assert response.json()["block"]["sub_block"]["in_use_by_ids"]


def test_other_subscriptions(test_client, generic_subscription_2, generic_product_type_2):
    _, GenericProductTwo = generic_product_type_2
    response = test_client.get(URL("api/subscriptions/instance/other_subscriptions/") / uuid4())
    assert response.status_code == HTTPStatus.NOT_FOUND

    subscription = GenericProductTwo.from_subscription(generic_subscription_2)
    response = test_client.get(
        URL("api/subscriptions/instance/other_subscriptions/") / subscription.pb_3.subscription_instance_id
    )
    assert response.status_code == HTTPStatus.OK
    assert len(response.json()) == 0


def test_set_in_sync(seed, test_client):
    subscription_id = IP_PREFIX_SUBSCRIPTION_ID
    unsync(subscription_id)
    db.session.commit()

    response = test_client.put(f"/api/subscriptions/{subscription_id}/set_in_sync")
    assert response.status_code == HTTPStatus.OK

    subscription = SubscriptionTable.query.get(subscription_id)
    assert subscription.insync


def _create_failed_process(subscription_id):
    process_id = uuid4()

    process = ProcessTable(
        process_id=process_id,
        workflow_name="validate_ip_prefix",
        last_status=ProcessStatus.FAILED,
        last_step="Verify references in NSO",
        assignee="NOC",
        is_task=False,
    )
    process_subscription = ProcessSubscriptionTable(process_id=process_id, subscription_id=subscription_id)

    db.session.add(process)
    db.session.add(process_subscription)

    db.session.commit()


def test_try_set_failed_task_in_sync(seed, test_client):
    subscription_id = IP_PREFIX_SUBSCRIPTION_ID
    unsync(IP_PREFIX_SUBSCRIPTION_ID)
    db.session.commit()

    _create_failed_process(subscription_id)

    response = test_client.put(f"/api/subscriptions/{subscription_id}/set_in_sync")
    assert response.status_code == HTTPStatus.UNPROCESSABLE_ENTITY

    subscription = SubscriptionTable.query.get(subscription_id)
    assert not subscription.insync


def test_product_block_paths(generic_subscription_1, generic_subscription_2):
    subscription_1 = SubscriptionModel.from_subscription(generic_subscription_1)
    subscription_2 = SubscriptionModel.from_subscription(generic_subscription_2)
    assert product_block_paths(subscription_1) == ["product", "pb_1", "pb_2"]
    assert product_block_paths(subscription_2) == ["product", "pb_3"]



# Copyright 2022-2023 SURF.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from typing import Any, Generic, TypeVar, Union

import strawberry

GenericType = TypeVar("GenericType")


@strawberry.type(description="Represents a paginated relationship between two entities")
class Connection(Generic[GenericType]):
    """Represents a paginated relationship between two entities.

    This pattern is used when the relationship itself has attributes.
    In a Facebook-based domain example, a friendship between two people
    would be a connection that might have a `friendshipStartTime`
    """

    page_info: "PageInfo"
    page: list["GenericType"]


@strawberry.federation.type(
    shareable=True, description="Pagination context to navigate objects with cursor-based pagination"
)
class PageInfo:
    """Pagination context to navigate objects with cursor-based pagination.

    Instead of classic offset pagination via `page` and `limit` parameters,
    here we have a cursor of the last object and we fetch items starting from that one

    Read more at:
        - https://graphql.org/learn/pagination/#pagination-and-edges
        - https://relay.dev/graphql/connections.htm
    """

    has_next_page: bool
    has_previous_page: bool
    start_cursor: Union[int, None]
    end_cursor: Union[int, None]
    total_items: Union[int, None]


@strawberry.type(description="An edge may contain additional information of the relationship")
class Edge(Generic[GenericType]):
    """An edge may contain additional information of the relationship. This is the trivial case."""

    node: GenericType
    cursor: str


EMPTY_PAGE: Connection[Any] = Connection(
    page=[],
    page_info=PageInfo(
        has_previous_page=False,
        has_next_page=False,
        start_cursor=0,
        end_cursor=-1,
        total_items=0,
    ),
)
